# encoding: utf-8

"""Unit test suite for `cr.cube.matrix.assembler` module."""

import numpy as np
import pytest

from cr.cube.cube import Cube
from cr.cube.dimension import Dimension, _Element, _OrderSpec, _Subtotal, _Subtotals
from cr.cube.enums import (
    COLLATION_METHOD as CM,
    DIMENSION_TYPE as DT,
    MARGINAL,
    MEASURE as M,
)
from cr.cube.matrix.assembler import (
    Assembler,
    _BaseOrderHelper,
    _BaseSortRowsByValueHelper,
    _ColumnOrderHelper,
    _RowOrderHelper,
    _SortRowsByBaseColumnHelper,
    _SortRowsByDerivedColumnHelper,
    _SortRowsByInsertedColumnHelper,
    _SortRowsByLabelHelper,
    _SortRowsByMarginalHelper,
)
from cr.cube.matrix.measure import (
    _BaseMarginal,
    _BaseSecondOrderMeasure,
    _ColumnComparableCounts,
    _ColumnIndex,
    _ColumnProportions,
    _ColumnShareSum,
    _ColumnStandardError,
    _ColumnUnweightedBases,
    _ColumnWeightedBases,
    _Means,
    _PopulationProportions,
    _PopulationStandardError,
    _MarginTableProportion,
    _MarginTableBase,
    _ProportionVariances,
    _Pvalues,
    _RowComparableCounts,
    _RowProportions,
    _RowStandardError,
    _RowShareSum,
    _RowUnweightedBases,
    _RowWeightedBases,
    _ScaleMean,
    _ScaleMeanStddev,
    _ScaleMeanStderr,
    _ScaleMedian,
    SecondOrderMeasures,
    _StdDev,
    _MarginWeightedBase,
    _Sums,
    _TotalShareSum,
    _TableProportions,
    _TableStandardError,
    _TableUnweightedBases,
    _TableBase,
    _TableBasesRange,
    _TableWeightedBases,
    _MarginUnweightedBase,
    _UnweightedCounts,
    _WeightedCounts,
    _Zscores,
)

from ...unitutil import (
    ANY,
    class_mock,
    instance_mock,
    method_mock,
    property_mock,
)


class DescribeAssembler:
    """Unit test suite for `cr.cube.matrix.assembler.Assembler` object."""

    @pytest.mark.parametrize(
        "measure_prop_name, MeasureCls",
        (
            ("column_comparable_counts", _ColumnComparableCounts),
            ("column_index", _ColumnIndex),
            ("column_proportions", _ColumnProportions),
            ("column_proportion_variances", _ProportionVariances),
            ("column_share_sum", _ColumnShareSum),
            ("column_std_err", _ColumnStandardError),
            ("column_unweighted_bases", _ColumnUnweightedBases),
            ("column_weighted_bases", _ColumnWeightedBases),
            ("means", _Means),
            ("population_proportions", _PopulationProportions),
            ("population_std_err", _PopulationStandardError),
            ("pvalues", _Pvalues),
            ("row_comparable_counts", _RowComparableCounts),
            ("row_proportions", _RowProportions),
            ("row_proportion_variances", _ProportionVariances),
            ("row_share_sum", _RowShareSum),
            ("row_std_err", _RowStandardError),
            ("row_unweighted_bases", _RowUnweightedBases),
            ("row_weighted_bases", _RowWeightedBases),
            ("stddev", _StdDev),
            ("sums", _Sums),
            ("table_proportions", _TableProportions),
            ("table_proportion_variances", _ProportionVariances),
            ("table_std_err", _TableStandardError),
            ("table_unweighted_bases", _TableUnweightedBases),
            ("table_weighted_bases", _TableWeightedBases),
            ("total_share_sum", _TotalShareSum),
            ("weighted_counts", _WeightedCounts),
            ("unweighted_counts", _UnweightedCounts),
            ("zscores", _Zscores),
        ),
    )
    def it_assembles_various_measures(
        self,
        request,
        _measures_prop_,
        second_order_measures_,
        _assemble_matrix_,
        measure_prop_name,
        MeasureCls,
    ):
        _measures_prop_.return_value = second_order_measures_
        setattr(
            second_order_measures_,
            measure_prop_name,
            instance_mock(request, MeasureCls, blocks=[["A", "B"], ["C", "D"]]),
        )
        _assemble_matrix_.return_value = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        assembler = Assembler(None, None, None)

        value = getattr(assembler, measure_prop_name)

        _assemble_matrix_.assert_called_once_with(assembler, [["A", "B"], ["C", "D"]])
        assert value == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

    @pytest.mark.parametrize(
        "measure_prop_name, MeasureCls",
        (
            ("rows_scale_mean", _ScaleMean),
            ("columns_scale_mean", _ScaleMean),
            ("rows_scale_mean_stddev", _ScaleMeanStddev),
            ("columns_scale_mean_stddev", _ScaleMeanStddev),
            ("rows_scale_mean_stderr", _ScaleMeanStderr),
            ("columns_scale_mean_stderr", _ScaleMeanStderr),
            ("rows_scale_median", _ScaleMedian),
            ("columns_scale_median", _ScaleMedian),
        ),
    )
    def it_assembles_various_marginals(
        self,
        request,
        _assemble_marginal_,
        _measures_prop_,
        second_order_measures_,
        measure_prop_name,
        MeasureCls,
    ):
        _assemble_marginal_.return_value = [[1, 2, 3], [4, 5, 6]]
        _measures_prop_.return_value = second_order_measures_
        measure_ = instance_mock(request, MeasureCls)
        setattr(second_order_measures_, measure_prop_name, measure_)
        assembler = Assembler(None, None, None)

        value = getattr(assembler, measure_prop_name)

        _assemble_marginal_.assert_called_once_with(assembler, measure_)
        assert value == [[1, 2, 3], [4, 5, 6]]

    def it_knows_the_column_labels(
        self,
        _columns_dimension_prop_,
        dimension_,
        _column_order_prop_,
        _dimension_labels_,
    ):
        _columns_dimension_prop_.return_value = dimension_
        _column_order_prop_.return_value = [0, 1, 2]
        _dimension_labels_.return_value = np.array(["Alpha", "Baker", "Charlie"])
        assembler = Assembler(None, None, None)

        column_labels = assembler.column_labels

        _dimension_labels_.assert_called_once_with(assembler, dimension_, [0, 1, 2])
        assert column_labels.tolist() == ["Alpha", "Baker", "Charlie"]

    def it_provides_a_1D_columns_base_for_an_X_CAT_cube_result(
        self,
        _measures_prop_,
        second_order_measures_,
        _assemble_marginal_,
        margin_unweighted_base_,
    ):
        margin_unweighted_base_.is_defined = True
        second_order_measures_.columns_unweighted_base = margin_unweighted_base_
        _assemble_marginal_.return_value = [[1, 2, 3], [4, 5, 6]]
        _measures_prop_.return_value = second_order_measures_
        assembler = Assembler(None, None, None)

        rows_base = assembler.columns_base

        _assemble_marginal_.assert_called_once_with(assembler, margin_unweighted_base_)
        assert rows_base == [[1, 2, 3], [4, 5, 6]]

    def but_it_provides_a_2D_columns_base_for_an_X_MR_cube_result(
        self,
        request,
        _measures_prop_,
        second_order_measures_,
        margin_unweighted_base_,
    ):
        margin_unweighted_base_.is_defined = False
        second_order_measures_.columns_unweighted_base = margin_unweighted_base_
        _measures_prop_.return_value = second_order_measures_
        property_mock(
            request,
            Assembler,
            "column_unweighted_bases",
            return_value=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        )
        assembler = Assembler(None, None, None)

        rows_base = assembler.columns_base

        assert rows_base == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

    def it_knows_the_columns_dimension_numeric_values(
        self, request, _columns_dimension_prop_, dimension_, _column_order_prop_
    ):
        _columns_dimension_prop_.return_value = dimension_
        dimension_.valid_elements = tuple(
            instance_mock(request, _Element, numeric_value=numeric_value)
            for numeric_value in (1, 2, 3)
        )
        _column_order_prop_.return_value = [2, -1, 0, -2]
        assembler = Assembler(None, None, None)

        np.testing.assert_almost_equal(
            assembler.columns_dimension_numeric_values,
            [3.0, np.nan, 1.0, np.nan],
        )

    def it_provides_a_1D_columns_margin_for_a_CAT_X_cube_result(
        self,
        _rows_dimension_prop_,
        dimension_,
        _measures_prop_,
        _assemble_marginal_,
        second_order_measures_,
        margin_weighted_base_,
    ):
        _rows_dimension_prop_.return_value = dimension_
        dimension_.dimension_type = DT.CAT
        _assemble_marginal_.return_value = [[1, 2, 3], [4, 5, 6]]
        _measures_prop_.return_value = second_order_measures_
        second_order_measures_.columns_weighted_base = margin_weighted_base_
        assembler = Assembler(None, None, None)

        columns_margin = assembler.columns_margin

        _assemble_marginal_.assert_called_once_with(assembler, margin_weighted_base_)
        assert columns_margin == [[1, 2, 3], [4, 5, 6]]

    def but_it_provides_a_2D_columns_margin_for_an_MR_X_cube_result(
        self,
        request,
        _measures_prop_,
        second_order_measures_,
        margin_weighted_base_,
    ):
        property_mock(
            request, Assembler, "column_weighted_bases", return_value=[[1, 2], [3, 4]]
        )
        _measures_prop_.return_value = second_order_measures_
        second_order_measures_.columns_weighted_base = margin_weighted_base_
        margin_weighted_base_.is_defined = False
        assembler = Assembler(None, None, None)

        columns_margin = assembler.columns_margin

        assert columns_margin == [[1, 2], [3, 4]]

    def it_provides_a_1D_columns_margin_proportion_for_a_CAT_X_cube_result(
        self,
        _rows_dimension_prop_,
        dimension_,
        _assemble_marginal_,
        _measures_prop_,
        margin_table_proportion_,
        second_order_measures_,
    ):
        margin_table_proportion_.is_defined = True
        _rows_dimension_prop_.return_value = dimension_
        dimension_.dimension_type = DT.CAT
        _assemble_marginal_.return_value = [[1, 2, 3], [4, 5, 6]]
        _measures_prop_.return_value = second_order_measures_
        second_order_measures_.columns_table_proportion = margin_table_proportion_
        assembler = Assembler(None, None, None)

        columns_margin_proportion = assembler.columns_margin_proportion

        _assemble_marginal_.assert_called_once_with(assembler, margin_table_proportion_)
        assert columns_margin_proportion == [[1, 2, 3], [4, 5, 6]]

    def but_it_provides_a_2D_columns_margin_proportion_for_an_MR_X_cube_result(
        self,
        request,
        dimensions_,
        _measures_prop_,
        second_order_measures_,
        margin_table_proportion_,
        SumSubtotals_,
        _assemble_matrix_,
    ):
        margin_table_proportion_.is_defined = False
        _measures_prop_.return_value = second_order_measures_
        second_order_measures_.columns_table_proportion = margin_table_proportion_
        property_mock(
            request,
            Assembler,
            "columns_margin",
            return_value=np.array([[1, 2], [3, 4]]),
        )
        property_mock(
            request,
            Assembler,
            "table_weighted_bases",
            return_value=np.array([4.0, 6.0]),
        )
        SumSubtotals_.blocks.return_value = [[[1], [2]], [[3], [4]]]
        _assemble_matrix_.return_value = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        assembler = Assembler(None, dimensions_, None)

        columns_margin_proportion = assembler.columns_margin_proportion

        SumSubtotals_.blocks.assert_called_once_with(ANY, dimensions_)
        assert SumSubtotals_.blocks.call_args.args[0].tolist() == [
            pytest.approx([1 / 4.0, 2 / 6.0]),
            pytest.approx([3 / 4.0, 4 / 6.0]),
        ]
        _assemble_matrix_.assert_called_once_with(assembler, [[[1], [2]], [[3], [4]]])
        assert columns_margin_proportion == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

    def it_knows_the_inserted_column_idxs(self, _column_order_prop_):
        _column_order_prop_.return_value = [2, -1, 0, -2]
        assert Assembler(None, None, None).inserted_column_idxs == (1, 3)

    def it_knows_the_inserted_row_idxs(self, _row_order_prop_):
        _row_order_prop_.return_value = [0, 1, -2, 2, -1, 3]
        assert Assembler(None, None, None).inserted_row_idxs == (2, 4)

    def it_knows_the_means(self, request, dimensions_):
        property_mock(
            request, Assembler, "means", return_value=np.array([1.2, 1.34, 3.3])
        )
        assembler = Assembler(None, dimensions_, None)

        assert assembler.means == pytest.approx([1.2, 1.34, 3.3])

    def it_knows_the_sum(self, request, dimensions_):
        property_mock(request, Assembler, "sums", return_value=np.array([4, 5, 6]))
        assembler = Assembler(None, dimensions_, None)

        assert assembler.sums == pytest.approx([4, 5, 6])

    def it_knows_the_row_labels(
        self,
        _rows_dimension_prop_,
        dimension_,
        _row_order_prop_,
        _dimension_labels_,
    ):
        _rows_dimension_prop_.return_value = dimension_
        _row_order_prop_.return_value = [0, 1, 2]
        _dimension_labels_.return_value = np.array(["Alpha", "Baker", "Charlie"])
        assembler = Assembler(None, None, None)

        row_labels = assembler.row_labels

        _dimension_labels_.assert_called_once_with(assembler, dimension_, [0, 1, 2])
        assert row_labels.tolist() == ["Alpha", "Baker", "Charlie"]

    def it_provides_a_1D_rows_base_for_an_X_CAT_cube_result(
        self,
        _measures_prop_,
        second_order_measures_,
        _assemble_marginal_,
        margin_unweighted_base_,
    ):
        margin_unweighted_base_.is_defined = True
        second_order_measures_.rows_unweighted_base = margin_unweighted_base_
        _assemble_marginal_.return_value = [[1, 2, 3], [4, 5, 6]]
        _measures_prop_.return_value = second_order_measures_
        assembler = Assembler(None, None, None)

        rows_base = assembler.rows_base

        _assemble_marginal_.assert_called_once_with(assembler, margin_unweighted_base_)
        assert rows_base == [[1, 2, 3], [4, 5, 6]]

    def but_it_provides_a_2D_rows_base_for_an_X_MR_cube_result(
        self,
        request,
        _measures_prop_,
        second_order_measures_,
        margin_unweighted_base_,
    ):
        margin_unweighted_base_.is_defined = False
        second_order_measures_.rows_unweighted_base = margin_unweighted_base_
        _measures_prop_.return_value = second_order_measures_
        property_mock(
            request,
            Assembler,
            "row_unweighted_bases",
            return_value=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        )
        assembler = Assembler(None, None, None)

        rows_base = assembler.rows_base

        assert rows_base == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

    @pytest.mark.parametrize(
        "order, expected_fills",
        (
            ((2, -2, 0, -1), ("#f00ba5", "STF1", "#000000", "STF2")),
            ((0, 1, 2, -2, -1), ("#000000", "#111111", "#f00ba5", "STF1", "STF2")),
            ((-2, -1, 0, 1, 2), ("STF1", "STF2", "#000000", "#111111", "#f00ba5")),
            ((-1, -2, 2, 1, 0), ("STF2", "STF1", "#f00ba5", "#111111", "#000000")),
        ),
    )
    def it_knows_the_rows_dimension_fills(
        self,
        request,
        _rows_dimension_prop_,
        dimension_,
        _row_order_prop_,
        order,
        expected_fills,
    ):
        element_fills = ("#000000", "#111111", "#f00ba5")
        subtotal_fills = ("STF1", "STF2")
        _rows_dimension_prop_.return_value = dimension_
        dimension_.valid_elements = tuple(
            instance_mock(request, _Element, fill=fill) for fill in element_fills
        )
        dimension_.subtotals = tuple(
            instance_mock(request, _Subtotal, fill=fill) for fill in subtotal_fills
        )
        _row_order_prop_.return_value = order
        assembler = Assembler(None, None, None)

        assert assembler.rows_dimension_fills == expected_fills

    def it_knows_the_rows_dimension_numeric_values(
        self, request, _rows_dimension_prop_, dimension_, _row_order_prop_
    ):
        _rows_dimension_prop_.return_value = dimension_
        dimension_.valid_elements = tuple(
            instance_mock(request, _Element, numeric_value=numeric_value)
            for numeric_value in (1, 2, 3)
        )
        _row_order_prop_.return_value = [2, -1, 0, -2]
        assembler = Assembler(None, None, None)

        np.testing.assert_almost_equal(
            assembler.rows_dimension_numeric_values,
            [3.0, np.nan, 1.0, np.nan],
        )

    def it_provides_a_1D_rows_margin_for_an_X_CAT_cube_result(
        self,
        _columns_dimension_prop_,
        dimension_,
        _measures_prop_,
        _assemble_marginal_,
        second_order_measures_,
        margin_weighted_base_,
    ):
        _columns_dimension_prop_.return_value = dimension_
        dimension_.dimension_type = DT.CAT
        _assemble_marginal_.return_value = [[1, 2, 3], [4, 5, 6]]
        _measures_prop_.return_value = second_order_measures_
        second_order_measures_.rows_weighted_base = margin_weighted_base_
        assembler = Assembler(None, None, None)

        rows_margin = assembler.rows_margin

        _assemble_marginal_.assert_called_once_with(assembler, margin_weighted_base_)
        assert rows_margin == [[1, 2, 3], [4, 5, 6]]

    def but_it_provides_a_2D_rows_margin_for_an_X_MR_cube_result(
        self,
        request,
        _measures_prop_,
        second_order_measures_,
        margin_weighted_base_,
    ):
        property_mock(
            request, Assembler, "row_weighted_bases", return_value=[[1, 2], [3, 4]]
        )
        _measures_prop_.return_value = second_order_measures_
        second_order_measures_.rows_weighted_base = margin_weighted_base_
        margin_weighted_base_.is_defined = False
        assembler = Assembler(None, None, None)

        rows_margin = assembler.rows_margin

        assert rows_margin == [[1, 2], [3, 4]]

    def it_provides_a_1D_rows_margin_proportion_for_an_X_CAT_cube_result(
        self,
        request,
        _columns_dimension_prop_,
        dimension_,
        _assemble_marginal_,
        _measures_prop_,
        second_order_measures_,
    ):
        _columns_dimension_prop_.return_value = dimension_
        dimension_.dimension_type = DT.CAT
        _assemble_marginal_.return_value = [[1, 2, 3], [4, 5, 6]]
        _measures_prop_.return_value = second_order_measures_
        measure_ = instance_mock(request, _MarginTableProportion)
        second_order_measures_.rows_table_proportion = measure_
        assembler = Assembler(None, None, None)

        rows_margin_proportion = assembler.rows_margin_proportion

        _assemble_marginal_.assert_called_once_with(assembler, measure_)
        assert rows_margin_proportion == [[1, 2, 3], [4, 5, 6]]

    def but_it_provides_a_2D_rows_margin_proportion_for_an_X_MR_cube_result(
        self,
        request,
        dimensions_,
        _measures_prop_,
        second_order_measures_,
        margin_table_proportion_,
        SumSubtotals_,
        _assemble_matrix_,
    ):
        margin_table_proportion_.is_defined = False
        _measures_prop_.return_value = second_order_measures_
        second_order_measures_.rows_table_proportion = margin_table_proportion_
        property_mock(
            request, Assembler, "rows_margin", return_value=np.array([[1, 2], [3, 4]])
        )
        property_mock(
            request,
            Assembler,
            "table_weighted_bases",
            return_value=np.array([4.0, 6.0]),
        )
        SumSubtotals_.blocks.return_value = [[[1], [2]], [[3], [4]]]
        _assemble_matrix_.return_value = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        assembler = Assembler(None, dimensions_, None)

        rows_margin_proportion = assembler.rows_margin_proportion

        SumSubtotals_.blocks.assert_called_once_with(ANY, dimensions_)
        assert SumSubtotals_.blocks.call_args.args[0].tolist() == [
            pytest.approx([1 / 4.0, 2 / 6.0]),
            pytest.approx([3 / 4.0, 4 / 6.0]),
        ]
        _assemble_matrix_.assert_called_once_with(assembler, [[[1], [2]], [[3], [4]]])
        assert rows_margin_proportion == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

    def it_knows_the_2D_table_base_of_an_ARRAY_X_ARRAY_matrix(
        self,
        request,
        _measures_prop_,
        second_order_measures_,
        table_unweighted_base_,
        rows_table_unweighted_base_,
        columns_table_unweighted_base_,
    ):
        table_unweighted_base_.is_defined = False
        rows_table_unweighted_base_.is_defined = False
        columns_table_unweighted_base_.is_defined = False
        second_order_measures_.table_unweighted_base = table_unweighted_base_
        second_order_measures_.rows_table_unweighted_base = rows_table_unweighted_base_
        second_order_measures_.columns_table_unweighted_base = (
            columns_table_unweighted_base_
        )
        _measures_prop_.return_value = second_order_measures_
        property_mock(
            request,
            Assembler,
            "table_unweighted_bases",
            return_value=[[5, 4, 6], [2, 1, 3]],
        )
        assembler = Assembler(None, None, None)

        assert assembler.table_base == [[5, 4, 6], [2, 1, 3]]

    def and_it_knows_the_1D_table_base_of_an_ARRAY_X_CAT_matrix(
        self,
        _measures_prop_,
        second_order_measures_,
        table_unweighted_base_,
        columns_table_unweighted_base_,
        _assemble_marginal_,
    ):
        table_unweighted_base_.is_defined = False
        columns_table_unweighted_base_.is_defined = True
        second_order_measures_.table_unweighted_base = table_unweighted_base_
        second_order_measures_.columns_table_unweighted_base = (
            columns_table_unweighted_base_
        )
        _measures_prop_.return_value = second_order_measures_
        _assemble_marginal_.return_value = [2, 1, 3]
        assembler = Assembler(None, None, None)

        table_base = assembler.table_base

        _assemble_marginal_.assert_called_once_with(
            assembler, columns_table_unweighted_base_
        )
        assert table_base == [2, 1, 3]

    def and_it_knows_the_1D_table_base_of_a_CAT_X_ARRAY_matrix(
        self,
        _measures_prop_,
        second_order_measures_,
        table_unweighted_base_,
        rows_table_unweighted_base_,
        columns_table_unweighted_base_,
        _assemble_marginal_,
    ):
        table_unweighted_base_.is_defined = False
        rows_table_unweighted_base_.is_defined = True
        columns_table_unweighted_base_.is_defined = False
        second_order_measures_.table_unweighted_base = table_unweighted_base_
        second_order_measures_.rows_table_unweighted_base = rows_table_unweighted_base_
        second_order_measures_.columns_table_unweighted_base = (
            columns_table_unweighted_base_
        )
        _measures_prop_.return_value = second_order_measures_
        _assemble_marginal_.return_value = [2, 1, 3]
        assembler = Assembler(None, None, None)

        table_base = assembler.table_base

        _assemble_marginal_.assert_called_once_with(
            assembler, rows_table_unweighted_base_
        )
        assert table_base == [2, 1, 3]

    def and_it_knows_the_scalar_table_base_of_a_CAT_X_CAT_matrix(
        self,
        _measures_prop_,
        second_order_measures_,
        table_unweighted_base_,
    ):
        table_unweighted_base_.is_defined = True
        table_unweighted_base_.value = 4242
        second_order_measures_.table_unweighted_base = table_unweighted_base_
        _measures_prop_.return_value = second_order_measures_
        assembler = Assembler(None, None, None)

        assert assembler.table_base == 4242

    def it_knows_the_2D_table_margin_of_an_ARRAY_X_ARRAY_matrix(
        self,
        request,
        _measures_prop_,
        second_order_measures_,
        table_weighted_base_,
        rows_table_weighted_base_,
        columns_table_weighted_base_,
    ):
        table_weighted_base_.is_defined = False
        rows_table_weighted_base_.is_defined = False
        columns_table_weighted_base_.is_defined = False
        second_order_measures_.table_weighted_base = table_weighted_base_
        second_order_measures_.rows_table_weighted_base = rows_table_weighted_base_
        second_order_measures_.columns_table_weighted_base = (
            columns_table_weighted_base_
        )
        _measures_prop_.return_value = second_order_measures_
        property_mock(
            request,
            Assembler,
            "table_weighted_bases",
            return_value=[[5, 4, 6], [2, 1, 3]],
        )
        assembler = Assembler(None, None, None)

        assert assembler.table_margin == [[5, 4, 6], [2, 1, 3]]

    def and_it_knows_the_1D_table_margin_of_an_ARRAY_X_CAT_matrix(
        self,
        _measures_prop_,
        second_order_measures_,
        table_weighted_base_,
        columns_table_weighted_base_,
        _assemble_marginal_,
    ):
        table_weighted_base_.is_defined = False
        columns_table_weighted_base_.is_defined = True
        second_order_measures_.table_weighted_base = table_weighted_base_
        second_order_measures_.columns_table_weighted_base = (
            columns_table_weighted_base_
        )
        _measures_prop_.return_value = second_order_measures_
        _assemble_marginal_.return_value = [2, 1, 3]
        assembler = Assembler(None, None, None)

        table_margin = assembler.table_margin

        _assemble_marginal_.assert_called_once_with(
            assembler, columns_table_weighted_base_
        )
        assert table_margin == [2, 1, 3]

    def and_it_knows_the_1D_table_margin_of_a_CAT_X_ARRAY_matrix(
        self,
        _measures_prop_,
        second_order_measures_,
        table_weighted_base_,
        rows_table_weighted_base_,
        columns_table_weighted_base_,
        _assemble_marginal_,
    ):
        table_weighted_base_.is_defined = False
        rows_table_weighted_base_.is_defined = True
        columns_table_weighted_base_.is_defined = False
        second_order_measures_.table_weighted_base = table_weighted_base_
        second_order_measures_.rows_table_weighted_base = rows_table_weighted_base_
        second_order_measures_.columns_table_weighted_base = (
            columns_table_weighted_base_
        )
        _measures_prop_.return_value = second_order_measures_
        _assemble_marginal_.return_value = [2, 1, 3]
        assembler = Assembler(None, None, None)

        table_margin = assembler.table_margin

        _assemble_marginal_.assert_called_once_with(
            assembler, rows_table_weighted_base_
        )
        assert table_margin == [2, 1, 3]

    def and_it_knows_the_scalar_table_margin_of_a_CAT_X_CAT_matrix(
        self,
        _measures_prop_,
        second_order_measures_,
        table_weighted_base_,
    ):
        table_weighted_base_.is_defined = True
        table_weighted_base_.value = 4242
        second_order_measures_.table_weighted_base = table_weighted_base_
        _measures_prop_.return_value = second_order_measures_
        assembler = Assembler(None, None, None)

        assert assembler.table_margin == 4242

    def it_knows_the_table_base_range(
        self,
        request,
        _measures_prop_,
        second_order_measures_,
    ):
        _measures_prop_.return_value = second_order_measures_
        measure_ = instance_mock(request, _TableBasesRange, value=42)
        second_order_measures_.table_unweighted_bases_range = measure_
        assembler = Assembler(None, None, None)

        assert assembler.table_base_range == 42

    def it_knows_the_table_margin_range(
        self,
        request,
        _measures_prop_,
        second_order_measures_,
    ):
        _measures_prop_.return_value = second_order_measures_
        measure_ = instance_mock(request, _TableBasesRange, value=42)
        second_order_measures_.table_weighted_bases_range = measure_
        assembler = Assembler(None, None, None)

        assert assembler.table_margin_range == 42

    # === implementation methods/properties ===

    @pytest.mark.parametrize(
        "row_order, col_order, blocks, expected_value",
        (
            ([0, 1], [0, 1, 2], [[1, 2, 3], [4, 5, 6]], [[1, 2, 3], [4, 5, 6]]),
            ([0, 1], [2, 1, 0], [[1, 2, 3], [4, 5, 6]], [[3, 2, 1], [6, 5, 4]]),
            ([0, 1], [1, 0], [[1, 2, 3], [4, 5, 6]], [[2, 1], [5, 4]]),
            ([0, 1], [2, 0], [[1, 2, 3], [4, 5, 6]], [[3, 1], [6, 4]]),
            ([1, 0], [0, 1, 2], [[1, 2, 3], [4, 5, 6]], [[4, 5, 6], [1, 2, 3]]),
            ([0], [0, 1, 2], [[1, 2, 3], [4, 5, 6]], [[1, 2, 3]]),
            ([0], [0], [[1, 2, 3], [4, 5, 6]], [[1]]),
        ),
    )
    def it_can_assemble_a_matrix_from_blocks(
        self,
        _row_order_prop_,
        row_order,
        _column_order_prop_,
        col_order,
        blocks,
        expected_value,
    ):
        _row_order_prop_.return_value = row_order
        _column_order_prop_.return_value = col_order
        assembler = Assembler(None, None, None)

        assert assembler._assemble_matrix(blocks).tolist() == expected_value

    def it_can_assemble_a_vector(self, request):
        base_vector = np.array([1, 2, 3, 4])
        subtotals_ = tuple(
            instance_mock(
                request,
                _Subtotal,
                addend_idxs=np.array(addend_idxs, dtype=int),
                subtrahend_idxs=np.array(subtrahend_idxs, dtype=int),
            )
            for addend_idxs, subtrahend_idxs in (
                ((0, 1), ()),
                ((1, 2), ()),
                ((2, 3), ()),
            )
        )
        order = np.array([-3, 1, 0, -2, 3, 2, -1])
        assembler = Assembler(None, None, None)

        vector = assembler._assemble_vector(base_vector, subtotals_, order)

        assert vector.tolist() == [3, 2, 1, 5, 4, 3, 7]

    def it_knows_the_column_order_to_help(
        self,
        _BaseOrderHelper_,
        dimensions_,
        _measures_prop_,
        second_order_measures_,
    ):
        _measures_prop_.return_value = second_order_measures_
        _BaseOrderHelper_.column_display_order.return_value = np.array(
            [-1, 1, -2, 2, -3, 3]
        )
        assembler = Assembler(None, dimensions_, None)

        column_order = assembler._column_order

        _BaseOrderHelper_.column_display_order.assert_called_once_with(
            dimensions_, second_order_measures_
        )
        assert column_order.tolist() == [-1, 1, -2, 2, -3, 3]

    def it_provides_access_to_the_column_subtotals_to_help(
        self, _columns_dimension_prop_, dimension_, subtotals_
    ):
        _columns_dimension_prop_.return_value = dimension_
        dimension_.subtotals = subtotals_

        assert Assembler(None, None, None)._column_subtotals is subtotals_

    def it_provides_access_to_the_columns_dimension_to_help(self, dimension_):
        assembler = Assembler(None, (None, dimension_), None)
        assert assembler._columns_dimension is dimension_

    def it_assembles_the_dimension_labels_to_help(self, dimension_):
        dimension_.element_labels = ("Alpha", "Bravo", "Charlie", "Delta")
        dimension_.subtotal_labels = ("Top 2", "All")
        order = np.array([1, 3, -2, 2, -1])
        assembler = Assembler(None, None, None)

        labels = assembler._dimension_labels(dimension_, order)

        assert labels.tolist() == ["Bravo", "Delta", "Top 2", "Charlie", "All"]

    def it_constructs_its_measures_collaborator_object_to_help(
        self, request, cube_, dimensions_, second_order_measures_
    ):
        SecondOrderMeasures_ = class_mock(
            request,
            "cr.cube.matrix.assembler.SecondOrderMeasures",
            return_value=second_order_measures_,
        )
        assembler = Assembler(cube_, dimensions_, slice_idx=17)

        measures = assembler._measures

        SecondOrderMeasures_.assert_called_once_with(cube_, dimensions_, 17)
        assert measures == second_order_measures_

    def it_knows_the_row_order_to_help(
        self,
        _BaseOrderHelper_,
        dimensions_,
        _measures_prop_,
        second_order_measures_,
    ):
        _measures_prop_.return_value = second_order_measures_
        _BaseOrderHelper_.row_display_order.return_value = np.array(
            [-1, 1, -2, 2, -3, 3]
        )
        assembler = Assembler(None, dimensions_, None)

        row_order = assembler._row_order

        _BaseOrderHelper_.row_display_order.assert_called_once_with(
            dimensions_, second_order_measures_
        )
        assert row_order.tolist() == [-1, 1, -2, 2, -3, 3]

    def it_provides_access_to_the_row_subtotals_to_help(
        self, _rows_dimension_prop_, dimension_, subtotals_
    ):
        _rows_dimension_prop_.return_value = dimension_
        dimension_.subtotals = subtotals_

        assert Assembler(None, None, None)._row_subtotals is subtotals_

    def it_knows_its_rows_dimension_to_help(self, dimension_):
        assembler = Assembler(None, (dimension_, None), None)
        assert assembler._rows_dimension is dimension_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _assemble_marginal_(self, request):
        return method_mock(request, Assembler, "_assemble_marginal")

    @pytest.fixture
    def _assemble_matrix_(self, request):
        return method_mock(request, Assembler, "_assemble_matrix")

    @pytest.fixture
    def _BaseOrderHelper_(self, request):
        return class_mock(request, "cr.cube.matrix.assembler._BaseOrderHelper")

    @pytest.fixture
    def _column_order_prop_(self, request):
        return property_mock(request, Assembler, "_column_order")

    @pytest.fixture
    def _columns_dimension_prop_(self, request):
        return property_mock(request, Assembler, "_columns_dimension")

    @pytest.fixture
    def columns_table_unweighted_base_(self, request):
        return instance_mock(request, _MarginTableBase)

    @pytest.fixture
    def columns_table_weighted_base_(self, request):
        return instance_mock(request, _MarginTableBase)

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)

    @pytest.fixture
    def _dimension_labels_(self, request):
        return method_mock(request, Assembler, "_dimension_labels")

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))

    @pytest.fixture
    def margin_table_proportion_(self, request):
        return instance_mock(request, _MarginTableProportion)

    @pytest.fixture
    def margin_unweighted_base_(self, request):
        return instance_mock(request, _MarginUnweightedBase)

    @pytest.fixture
    def margin_weighted_base_(self, request):
        return instance_mock(request, _MarginWeightedBase)

    @pytest.fixture
    def _measures_prop_(self, request):
        return property_mock(request, Assembler, "_measures")

    @pytest.fixture
    def _row_order_prop_(self, request):
        return property_mock(request, Assembler, "_row_order")

    @pytest.fixture
    def _rows_dimension_prop_(self, request):
        return property_mock(request, Assembler, "_rows_dimension")

    @pytest.fixture
    def rows_table_unweighted_base_(self, request):
        return instance_mock(request, _MarginTableBase)

    @pytest.fixture
    def rows_table_weighted_base_(self, request):
        return instance_mock(request, _MarginTableBase)

    @pytest.fixture
    def second_order_measures_(self, request):
        return instance_mock(request, SecondOrderMeasures)

    @pytest.fixture
    def subtotals_(self, request):
        return instance_mock(request, _Subtotals)

    @pytest.fixture
    def SumSubtotals_(self, request):
        return class_mock(request, "cr.cube.matrix.assembler.SumSubtotals")

    @pytest.fixture
    def table_unweighted_base_(self, request):
        return instance_mock(request, _TableBase)

    @pytest.fixture
    def table_weighted_base_(self, request):
        return instance_mock(request, _TableBase)


class Describe_BaseOrderHelper:
    """Unit test suite for `cr.cube.matrix.assembler._BaseOrderHelper` object."""

    def it_dispatches_to_the_right_column_order_helper(
        self, request, dimensions_, second_order_measures_
    ):
        column_order_helper_ = instance_mock(
            request, _ColumnOrderHelper, _display_order=np.array([-2, 1, -1, 2])
        )
        _ColumnOrderHelper_ = class_mock(
            request,
            "cr.cube.matrix.assembler._ColumnOrderHelper",
            return_value=column_order_helper_,
        )

        column_order = _BaseOrderHelper.column_display_order(
            dimensions_, second_order_measures_
        )

        _ColumnOrderHelper_.assert_called_once_with(dimensions_, second_order_measures_)
        assert column_order.tolist() == [-2, 1, -1, 2]

    @pytest.mark.parametrize(
        "collation_method, HelperCls",
        (
            (CM.EXPLICIT_ORDER, _RowOrderHelper),
            (CM.LABEL, _SortRowsByLabelHelper),
            (CM.MARGINAL, _SortRowsByMarginalHelper),
            (CM.OPPOSING_ELEMENT, _SortRowsByBaseColumnHelper),
            (CM.PAYLOAD_ORDER, _RowOrderHelper),
        ),
    )
    def it_dispatches_to_the_right_row_order_helper(
        self, request, dimensions_, second_order_measures_, collation_method, HelperCls
    ):
        dimensions_[0].order_spec = instance_mock(
            request, _OrderSpec, collation_method=collation_method
        )
        helper_ = instance_mock(
            request, HelperCls, _display_order=np.array([-1, 1, -2, 2])
        )
        HelperCls_ = class_mock(
            request,
            "cr.cube.matrix.assembler.%s" % HelperCls.__name__,
            return_value=helper_,
        )

        row_order = _BaseOrderHelper.row_display_order(
            dimensions_, second_order_measures_
        )

        HelperCls_.assert_called_once_with(dimensions_, second_order_measures_)
        assert row_order.tolist() == [-1, 1, -2, 2]

    def it_provides_access_to_the_columns_dimension_to_help(self, dimension_):
        order_helper = _BaseOrderHelper((None, dimension_), None)
        assert order_helper._columns_dimension is dimension_

    @pytest.mark.parametrize(
        "prune_subtotals, order, expected_value",
        (
            (True, (-1, 1, -2, 2, -3, 3), [1, 2, 3]),
            (True, (1, 2, 3), [1, 2, 3]),
            (False, (-1, 1, -2, 2, -3, 3), [-1, 1, -2, 2, -3, 3]),
        ),
    )
    def it_post_processes_the_display_order_to_help(
        self, request, prune_subtotals, order, expected_value
    ):
        property_mock(
            request, _BaseOrderHelper, "_prune_subtotals", return_value=prune_subtotals
        )
        property_mock(request, _BaseOrderHelper, "_order", return_value=order)
        order_helper = _BaseOrderHelper(None, None)

        display_order = order_helper._display_order

        assert display_order.tolist() == expected_value

    @pytest.mark.parametrize(
        "base, expected_value",
        (
            ([False, False, False], ()),
            ([False, True, False], (1,)),
            ([True, True, True], (0, 1, 2)),
        ),
    )
    def it_knows_its_empty_column_idxs_to_help(
        self, second_order_measures_, base, expected_value
    ):
        second_order_measures_.columns_pruning_mask = np.array(base)
        order_helper = _BaseOrderHelper(None, second_order_measures_)

        assert order_helper._empty_column_idxs == expected_value

    @pytest.mark.parametrize(
        "base, expected_value",
        (
            ([False, False, False], ()),
            ([False, True, False], (1,)),
            ([True, True, True], (0, 1, 2)),
        ),
    )
    def it_knows_its_empty_row_idxs_to_help(
        self, second_order_measures_, base, expected_value
    ):
        second_order_measures_.rows_pruning_mask = np.array(base)
        order_helper = _BaseOrderHelper(None, second_order_measures_)

        assert order_helper._empty_row_idxs == expected_value

    @pytest.mark.parametrize(
        "json_name, internal_name",
        (
            ("col_base_unweighted", "column_unweighted_bases"),
            ("col_base_weighted", "column_weighted_bases"),
            ("col_index", "column_index"),
            ("col_percent", "column_proportions"),
            ("col_percent_moe", "column_std_err"),
            ("col_share_sum", "column_share_sum"),
            ("col_std_dev", "column_proportion_variances"),
            ("col_std_err", "column_std_err"),
            ("count_unweighted", "unweighted_counts"),
            ("count_weighted", "weighted_counts"),
            ("mean", "means"),
            ("population", "population_proportions"),
            ("population_moe", "population_std_err"),
            ("p_value", "pvalues"),
            ("row_base_unweighted", "row_unweighted_bases"),
            ("row_base_weighted", "row_weighted_bases"),
            ("row_percent", "row_proportions"),
            ("row_percent_moe", "row_std_err"),
            ("row_share_sum", "row_share_sum"),
            ("row_std_dev", "row_proportion_variances"),
            ("row_std_err", "row_std_err"),
            ("stddev", "stddev"),
            ("sum", "sums"),
            ("table_base_unweighted", "table_unweighted_bases"),
            ("table_base_weighted", "table_weighted_bases"),
            ("table_percent", "table_proportions"),
            ("table_percent_moe", "table_std_err"),
            ("table_std_dev", "table_proportion_variances"),
            ("table_std_err", "table_std_err"),
            ("total_share_sum", "total_share_sum"),
            ("valid_count_unweighted", "unweighted_counts"),
            ("valid_count_weighted", "weighted_counts"),
            ("z_score", "zscores"),
        ),
    )
    def it_retrieves_the_measure_object_to_help(
        self, request, second_order_measures_, json_name, internal_name
    ):
        property_mock(
            request,
            _BaseOrderHelper,
            "_order_spec",
            return_value=instance_mock(request, _OrderSpec, measure=M(json_name)),
        )
        measure_ = instance_mock(request, _BaseSecondOrderMeasure)
        setattr(second_order_measures_, internal_name, measure_)
        order_helper = _BaseOrderHelper(None, second_order_measures_)

        assert order_helper._measure is measure_

    def it_provides_access_to_the_rows_dimension_to_help(self, dimension_):
        order_helper = _BaseOrderHelper((dimension_, None), None)
        assert order_helper._rows_dimension is dimension_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))

    @pytest.fixture
    def second_order_measures_(self, request):
        return instance_mock(request, SecondOrderMeasures)


class Describe_ColumnOrderHelper:
    """Unit test suite for `cr.cube.matrix.assembler._ColumnOrderHelper` object."""

    @pytest.mark.parametrize(
        "collation_method, collator_class_name",
        (
            (CM.PAYLOAD_ORDER, "PayloadOrderCollator"),
            (CM.EXPLICIT_ORDER, "ExplicitOrderCollator"),
        ),
    )
    def it_computes_the_order_of_a_columns_dimension_to_help(
        self, request, dimension_, collation_method, collator_class_name
    ):
        property_mock(
            request, _ColumnOrderHelper, "_columns_dimension", return_value=dimension_
        )
        dimension_.order_spec = instance_mock(
            request, _OrderSpec, collation_method=collation_method
        )
        CollatorCls_ = class_mock(
            request, "cr.cube.matrix.assembler.%s" % collator_class_name
        )
        CollatorCls_.display_order.return_value = (3, -1, 5, 1, -2)
        property_mock(
            request, _ColumnOrderHelper, "_empty_column_idxs", return_value=(1, 3)
        )
        order_helper = _ColumnOrderHelper(None, None)

        order = order_helper._order

        CollatorCls_.display_order.assert_called_once_with(dimension_, (1, 3))
        assert order == (3, -1, 5, 1, -2)

    @pytest.mark.parametrize(
        "prune, empty_row_idxs, expected_value",
        (
            (False, None, False),
            (True, (3,), False),
            (True, (0, 1, 2), True),
        ),
    )
    def it_knows_whether_to_prune_the_subtotal_columns_to_help(
        self, request, dimension_, prune, empty_row_idxs, expected_value
    ):
        property_mock(
            request, _ColumnOrderHelper, "_rows_dimension", return_value=dimension_
        )
        property_mock(
            request,
            _ColumnOrderHelper,
            "_empty_row_idxs",
            return_value=empty_row_idxs,
        )
        dimension_.prune = prune
        dimension_.element_ids = (1, 2, 3)
        order_helper = _ColumnOrderHelper(None, None)

        assert order_helper._prune_subtotals is expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)


class Describe_RowOrderHelper:
    """Unit test suite for `cr.cube.matrix.assembler._RowOrderHelper` object."""

    @pytest.mark.parametrize(
        "collation_method, collator_class_name",
        (
            (CM.PAYLOAD_ORDER, "PayloadOrderCollator"),
            (CM.EXPLICIT_ORDER, "ExplicitOrderCollator"),
        ),
    )
    def it_computes_the_order_of_a_rows_dimension_to_help(
        self, request, dimension_, collation_method, collator_class_name
    ):
        property_mock(
            request, _RowOrderHelper, "_rows_dimension", return_value=dimension_
        )
        dimension_.order_spec = instance_mock(
            request, _OrderSpec, collation_method=collation_method
        )
        CollatorCls_ = class_mock(
            request, "cr.cube.matrix.assembler.%s" % collator_class_name
        )
        CollatorCls_.display_order.return_value = (1, -2, 3, 5, -1)
        property_mock(
            request, _RowOrderHelper, "_empty_row_idxs", return_value=(2, 4, 6)
        )
        order_helper = _RowOrderHelper(None, None)

        order = order_helper._order

        CollatorCls_.display_order.assert_called_once_with(dimension_, (2, 4, 6))
        assert order == (1, -2, 3, 5, -1)

    def it_provides_access_to_the_order_spec_to_help(self, request, dimension_):
        property_mock(
            request, _RowOrderHelper, "_rows_dimension", return_value=dimension_
        )
        order_spec_ = instance_mock(request, _OrderSpec)
        dimension_.order_spec = order_spec_
        order_helper = _SortRowsByBaseColumnHelper(None, None)

        assert order_helper._order_spec is order_spec_

    @pytest.mark.parametrize(
        "prune, empty_column_idxs, expected_value",
        (
            (False, None, False),
            (True, (3,), False),
            (True, (0, 1, 2), True),
        ),
    )
    def it_knows_whether_to_prune_the_subtotal_rows_to_help(
        self, request, dimension_, prune, empty_column_idxs, expected_value
    ):
        property_mock(
            request, _RowOrderHelper, "_columns_dimension", return_value=dimension_
        )
        property_mock(
            request,
            _RowOrderHelper,
            "_empty_column_idxs",
            return_value=empty_column_idxs,
        )
        dimension_.prune = prune
        dimension_.element_ids = (1, 2, 3)
        order_helper = _RowOrderHelper(None, None)

        assert order_helper._prune_subtotals is expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)


class Describe_BaseSortRowsByValueHelper:
    """Unit test suite for `cr.cube.matrix.assembler._BaseSortRowsByValueHelper`."""

    def it_provides_the_order(
        self,
        SortByValueCollator_,
        _rows_dimension_prop_,
        _element_values_prop_,
        _subtotal_values_prop_,
        _empty_row_idxs_prop_,
    ):
        _BaseSortRowsByValueHelper(None, None)._order

        SortByValueCollator_.display_order.assert_called_once_with(
            _rows_dimension_prop_(),
            _element_values_prop_(),
            _subtotal_values_prop_(),
            _empty_row_idxs_prop_(),
        )

    def but_it_falls_back_to_payload_order_on_value_error(
        self,
        request,
        dimensions_,
        _element_values_prop_,
        _subtotal_values_prop_,
        _empty_row_idxs_prop_,
        SortByValueCollator_,
    ):
        _element_values_prop_.return_value = None
        _subtotal_values_prop_.return_value = None
        _empty_row_idxs_prop_.return_value = (4, 2)
        SortByValueCollator_.display_order.side_effect = ValueError
        PayloadOrderCollator_ = class_mock(
            request, "cr.cube.matrix.assembler.PayloadOrderCollator"
        )
        PayloadOrderCollator_.display_order.return_value = (1, 2, 3, 4)
        order_helper = _BaseSortRowsByValueHelper(dimensions_, None)

        order = order_helper._order

        PayloadOrderCollator_.display_order.assert_called_once_with(
            dimensions_[0], (4, 2)
        )
        assert order == (1, 2, 3, 4)

    # fixture components ---------------------------------------------

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))

    @pytest.fixture
    def _element_values_prop_(self, request):
        return property_mock(request, _BaseSortRowsByValueHelper, "_element_values")

    @pytest.fixture
    def _empty_row_idxs_prop_(self, request):
        return property_mock(request, _BaseSortRowsByValueHelper, "_empty_row_idxs")

    @pytest.fixture
    def _rows_dimension_prop_(self, request):
        return property_mock(request, _BaseSortRowsByValueHelper, "_rows_dimension")

    @pytest.fixture
    def SortByValueCollator_(self, request):
        return class_mock(request, "cr.cube.matrix.assembler.SortByValueCollator")

    @pytest.fixture
    def _subtotal_values_prop_(self, request):
        return property_mock(request, _BaseSortRowsByValueHelper, "_subtotal_values")


class Describe_SortRowsByBaseColumnHelper:
    """Unit test suite for `cr.cube.matrix.assembler._SortRowsByBaseColumnHelper`."""

    def it_derives_the_sort_column_idx_from_the_order_spec_to_help(
        self, _columns_dimension_prop_, dimension_, _order_spec_prop_, order_spec_
    ):
        _columns_dimension_prop_.return_value = dimension_
        dimension_.element_ids = (1, 2, 3, 4, 5)
        dimension_.translate_element_id.return_value = 3
        _order_spec_prop_.return_value = order_spec_
        order_spec_.element_id = "c"
        order_helper = _SortRowsByBaseColumnHelper(None, None)

        column_idx = order_helper._column_idx

        dimension_.translate_element_id.assert_called_once_with("c")
        assert column_idx == 2

    def it_extracts_the_element_values_to_help(
        self, _measure_prop_, measure_, _column_idx_prop_
    ):
        _measure_prop_.return_value = measure_
        measure_.blocks = [[np.arange(20).reshape(4, 5), None], [None, None]]
        _column_idx_prop_.return_value = 2
        order_helper = _SortRowsByBaseColumnHelper(None, None)

        assert order_helper._element_values.tolist() == [2, 7, 12, 17]

    def but_it_raises_when_an_unsupported_sort_by_value_measure_is_requested(
        self, _order_spec_prop_, order_spec_
    ):
        _order_spec_prop_.return_value = order_spec_
        order_spec_.measure = "foo"
        order_helper = _SortRowsByBaseColumnHelper(None, None)

        with pytest.raises(NotImplementedError) as e:
            order_helper._measure

        assert str(e.value) == ("sort-by-value for measure 'foo' is not yet supported")

    def it_extracts_the_subtotal_values_to_help(
        self, _measure_prop_, measure_, _column_idx_prop_
    ):
        _measure_prop_.return_value = measure_
        measure_.blocks = [[None, None], [np.arange(10, 101, 10).reshape(2, 5), None]]
        _column_idx_prop_.return_value = 2
        order_helper = _SortRowsByBaseColumnHelper(None, None)

        assert order_helper._subtotal_values.tolist() == [30, 80]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _column_idx_prop_(self, request):
        return property_mock(request, _SortRowsByBaseColumnHelper, "_column_idx")

    @pytest.fixture
    def _columns_dimension_prop_(self, request):
        return property_mock(request, _SortRowsByBaseColumnHelper, "_columns_dimension")

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)

    @pytest.fixture
    def measure_(self, request):
        return instance_mock(request, _BaseSecondOrderMeasure)

    @pytest.fixture
    def _measure_prop_(self, request):
        return property_mock(request, _SortRowsByBaseColumnHelper, "_measure")

    @pytest.fixture
    def order_spec_(self, request):
        return instance_mock(request, _OrderSpec)

    @pytest.fixture
    def _order_spec_prop_(self, request):
        return property_mock(request, _SortRowsByBaseColumnHelper, "_order_spec")


class Describe_SortRowsByDerivedColumnHelper:
    """Unit test suite for `cr.cube.matrix.assembler._SortRowsByDerivedColumnHelper`."""

    def it_derives_the_sort_column_idx_from_the_order_spec_to_help(
        self, _columns_dimension_prop_, dimension_, _order_spec_prop_, order_spec_
    ):
        _columns_dimension_prop_.return_value = dimension_
        dimension_.element_ids = (1, 2, 3, 4, 5)
        dimension_.translate_element_id.return_value = 3
        _order_spec_prop_.return_value = order_spec_
        order_spec_.insertion_id = "c"
        order_helper = _SortRowsByDerivedColumnHelper(None, None)

        column_idx = order_helper._column_idx

        dimension_.translate_element_id.assert_called_once_with("c")
        assert column_idx == 2

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _columns_dimension_prop_(self, request):
        return property_mock(request, _SortRowsByBaseColumnHelper, "_columns_dimension")

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)

    @pytest.fixture
    def order_spec_(self, request):
        return instance_mock(request, _OrderSpec)

    @pytest.fixture
    def _order_spec_prop_(self, request):
        return property_mock(request, _SortRowsByBaseColumnHelper, "_order_spec")


class Describe_SortRowsByInsertedColumnHelper:
    """Unit test suite for `cr.cube.matrix.assembler._SortRowsByInsertedColumnHelper`."""

    def it_extracts_the_element_values_to_help(
        self, _measure_prop_, measure_, _insertion_idx_prop_
    ):
        _measure_prop_.return_value = measure_
        measure_.blocks = [[None, np.arange(12).reshape(4, 3)], [None, None]]
        _insertion_idx_prop_.return_value = 1
        order_helper = _SortRowsByInsertedColumnHelper(None, None)

        assert order_helper._element_values.tolist() == [1, 4, 7, 10]

    def it_derives_the_sort_insertion_idx_from_the_order_spec_to_help(
        self, _columns_dimension_prop_, dimension_, _order_spec_prop_, order_spec_
    ):
        _columns_dimension_prop_.return_value = dimension_
        dimension_.insertion_ids = (1, 2, 3)
        _order_spec_prop_.return_value = order_spec_
        order_spec_.insertion_id = 2
        order_helper = _SortRowsByInsertedColumnHelper(None, None)

        assert order_helper._insertion_idx == 1

    def it_extracts_the_subtotal_values_to_help(
        self, _measure_prop_, measure_, _insertion_idx_prop_
    ):
        _measure_prop_.return_value = measure_
        measure_.blocks = [[None, None], [None, np.arange(10, 61, 10).reshape(2, 3)]]
        _insertion_idx_prop_.return_value = 1
        order_helper = _SortRowsByInsertedColumnHelper(None, None)

        assert order_helper._subtotal_values.tolist() == [20, 50]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _columns_dimension_prop_(self, request):
        return property_mock(
            request, _SortRowsByInsertedColumnHelper, "_columns_dimension"
        )

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)

    @pytest.fixture
    def _insertion_idx_prop_(self, request):
        return property_mock(request, _SortRowsByInsertedColumnHelper, "_insertion_idx")

    @pytest.fixture
    def measure_(self, request):
        return instance_mock(request, _BaseSecondOrderMeasure)

    @pytest.fixture
    def _measure_prop_(self, request):
        return property_mock(request, _SortRowsByInsertedColumnHelper, "_measure")

    @pytest.fixture
    def order_spec_(self, request):
        return instance_mock(request, _OrderSpec)

    @pytest.fixture
    def _order_spec_prop_(self, request):
        return property_mock(request, _SortRowsByInsertedColumnHelper, "_order_spec")


class Describe_SortRowsByLabelHelper:
    """Unit test suite for `cr.cube.matrix.assembler._SortRowsByLabelHelper`."""

    def it_provides_the_element_values_to_help(self, dimensions_):
        dimensions_[0].element_labels = ("c", "a", "b")

        assert _SortRowsByLabelHelper(dimensions_, None)._element_values.tolist() == [
            "c",
            "a",
            "b",
        ]

    def it_provides_the_subtotal_values_to_help(self, dimensions_):
        dimensions_[0].subtotal_labels = ("c", "a", "b")

        assert _SortRowsByLabelHelper(dimensions_, None)._subtotal_values.tolist() == [
            "c",
            "a",
            "b",
        ]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def dimensions_(self, request):
        return (instance_mock(request, Dimension), instance_mock(request, Dimension))


class Describe_SortRowsByMarginalHelper:
    """Unit test suite for `cr.cube.matrix.assembler._SortRowsByMarginalHelper`."""

    def it_provides_the_element_values_to_help(self, _marginal_prop_, marginal_):
        marginal_.blocks = ["a", "b"]
        _marginal_prop_.return_value = marginal_

        assert _SortRowsByMarginalHelper(None, None)._element_values == "a"

    @pytest.mark.parametrize(
        "marginal, marginal_prop_name",
        (
            (MARGINAL.BASE, "rows_unweighted_base"),
            (MARGINAL.MARGIN, "rows_weighted_base"),
            (MARGINAL.MARGIN_PROPORTION, "rows_table_proportion"),
            (MARGINAL.SCALE_MEAN, "rows_scale_mean"),
            (MARGINAL.SCALE_MEAN_STDDEV, "rows_scale_mean_stddev"),
            (MARGINAL.SCALE_MEAN_STDERR, "rows_scale_mean_stderr"),
            (MARGINAL.SCALE_MEDIAN, "rows_scale_median"),
        ),
    )
    def it_provides_the_marginal_to_help(
        self,
        second_order_measures_,
        marginal,
        marginal_prop_name,
        _order_spec_,
        _order_spec_prop_,
    ):
        setattr(second_order_measures_, marginal_prop_name, "foo")
        _order_spec_.marginal = marginal
        _order_spec_prop_.return_value = _order_spec_
        order_helper = _SortRowsByMarginalHelper(None, second_order_measures_)

        assert order_helper._marginal == "foo"

    def but_it_raises_on_unknown_marginal(
        self, second_order_measures_, _order_spec_, _order_spec_prop_
    ):
        _order_spec_.marginal = "bar"
        _order_spec_prop_.return_value = _order_spec_
        order_helper = _SortRowsByMarginalHelper(None, second_order_measures_)

        with pytest.raises(NotImplementedError) as e:
            order_helper._marginal

        assert str(e.value) == "sort-by-value for marginal 'bar' is not yet supported"

    def it_provides_the_subtotal_values_to_help(self, _marginal_prop_, marginal_):
        marginal_.blocks = ["a", "b"]
        _marginal_prop_.return_value = marginal_

        assert _SortRowsByMarginalHelper(None, None)._subtotal_values == "b"

    # fixture components ---------------------------------------------

    @pytest.fixture
    def marginal_(self, request):
        return instance_mock(request, _BaseMarginal)

    @pytest.fixture
    def _marginal_prop_(self, request):
        return property_mock(request, _SortRowsByMarginalHelper, "_marginal")

    @pytest.fixture
    def _order_spec_(self, request):
        return instance_mock(request, _OrderSpec)

    @pytest.fixture
    def _order_spec_prop_(self, request):
        return property_mock(request, _SortRowsByMarginalHelper, "_order_spec")

    @pytest.fixture
    def second_order_measures_(self, request):
        return instance_mock(request, SecondOrderMeasures)



# encoding: utf-8

import warnings

import numpy as np

from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.util import lazyproperty


class Smoother(object):
    """Base object class for Smoother variants."""

    @classmethod
    def factory(cls, dimension):
        """Returns appropriate Smoother object according to passed function.

        Raises an exception if the function is different from `one_sided_moving_avg`
        that at the moment is the only one implemented.
        """
        smoothing_dict = dimension.smoothing_dict
        function = smoothing_dict.get("function") or "one_sided_moving_avg"
        if function != "one_sided_moving_avg":
            raise NotImplementedError("Function {} is not available.".format(function))
        return _SingleSidedMovingAvgSmoother(
            smoothing_dict=smoothing_dict,
            dimension_type=dimension.dimension_type,
        )


class _SingleSidedMovingAvgSmoother(object):
    """Create and configure smoothing function for one-sided moving average."""

    def __init__(self, smoothing_dict, dimension_type):
        self._smoothing_dict = smoothing_dict
        self._dimension_type = dimension_type

    def smooth(self, values):
        """1D/2D float64 ndarray of smoothed values including additional nans.

        Given a series of numbers and a fixed subset size, the first element of the
        moving average is obtained by taking the average of the initial fixed subset
        of the number series. Then the subset is modified by `shifting forward` the
        values. A moving average is commonly used with time series data to smooth
        out short-term fluctuations and highlight longer-term trends or cycles.

        The below examples will show 1D and 2D array rolling mean calculations with
        window sizes of 2 and 3, respectively.

                                    [window = 2]
        ----------------------------------------------------------------------------
            x    |   smooth(x)                  x     |        smooth(x)
        ---------+--------------         -------------+------------------------
            1    |    NaN                 1  3  2  3  |   NaN  2.0  2.5  2.5
            2    |    1.5                 2  3  3  2  |   NaN  2.5  3.0  2.5
            3    |    2.5                 3  2  4  4  |   NaN  2.5  3.0  4.0
            4    |    3.5                 4  1  5  1  |   NaN  2.5  3.0  3.0

                                    [window = 3]
        ----------------------------------------------------------------------------
            x    |   smooth(x)                  x     |        smooth(x)
        ---------+--------------         -------------+------------------------
            1    |    NaN                 1  3  2  3  |   NaN  NaN   2.0  2.67
            2    |    NaN                 2  3  3  2  |   NaN  NaN  2.67  2.67
            3    |     2                  3  2  4  4  |   NaN  NaN   3.0  3.33
            4    |     3                  4  1  5  1  |   NaN  NaN  3.33  2.33

        This is performed just taking the average of the last 2 or 3 rows according
        to the window, all the way down the column.
        """
        window = self._window
        if not self._can_smooth(values):
            # return original values if smoothing cannot be performed
            return values
        smoothed_values = (
            np.array(tuple(np.convolve(values, np.ones(window), mode="valid") / window))
            if values.ndim == 1
            else np.array(
                [
                    tuple(np.convolve(v, np.ones(window), mode="valid") / window)
                    for v in values
                ]
            )
        )
        # offset between original values and smoothed values
        offset = [values.shape[-1] - smoothed_values.shape[-1]]
        additional_nans = np.full(list(values.shape[:-1]) + offset, np.nan)
        return np.concatenate([additional_nans, smoothed_values], axis=values.ndim - 1)

    def _can_smooth(self, base_values):
        """Returns True if the values can be smoothed, False otherwise.

        If the spec parameters describe valid smoothing and the column dimensions is a
        categorical date it returns True, False otherwise.
        """
        window = self._window
        if base_values.size == 0:
            # --- when base values are empty just return them. Cannot smooth an empty
            # --- array.
            return False
        if not self._dimension_type == DT.CAT_DATE:
            # --- smoothing can be performed only on categorical date dimension type
            orienation = "Row" if base_values.ndim == 1 else "Column"
            warnings.warn(
                "No smoothing performed. {} dimension must be a categorical "
                "date.".format(orienation)
            )
            return False
        # --- window cannot be wider than the number of periods in data
        # --- and it must be at least 2.
        if window > base_values.shape[-1] or window < 2:
            warnings.warn(
                "No smoothing performed. Smoothing window must be between 2 and the "
                "number of periods ({}), got {}".format(base_values.shape[-1], window),
            )
            return False
        return True

    @lazyproperty
    def _window(self):
        """Int representing the window for the moving avarage specified in the dict."""
        return self._smoothing_dict.get("window") or 2



# encoding: utf-8

"""The `Assembler` object provides the external interface for this module.

Its name derives from its role to "assemble" a finished 2D array ("matrix") for a
particular measure from the base measure values and inserted subtotals, to reorder the
rows and columns according to the dimension *order* transforms, and to hide rows and
columns that are either hidden by the user or "pruned" because they contain no
observations.
"""

import numpy as np

from cr.cube.collator import (
    ExplicitOrderCollator,
    PayloadOrderCollator,
    SortByValueCollator,
)
from cr.cube.enums import (
    COLLATION_METHOD as CM,
    DIMENSION_TYPE as DT,
    MARGINAL,
    MARGINAL_ORIENTATION as MO,
    MEASURE as M,
)
from cr.cube.matrix.measure import SecondOrderMeasures
from cr.cube.matrix.subtotals import SumSubtotals
from cr.cube.util import lazyproperty


class Assembler:
    """Provides measure and margin methods for a cube-slice.

    An assembled matrix is a 2D ndarray reflecting all ordering, insertion, and hiding
    transforms applied to the dimensions. An assembled margin is often a 1D ndarray
    which is similarly formed from inserted values, ordered, and value hiding applied.

    `cube` is the `cr.cube.Cube` object containing the data for this matrix. Note that
    not all the data in `cube` will necessarily be used by this matrix. When `cube` is
    more than 2-dimensional, it is "sliced" and each slice gets its own matrix (and
    `_Slice` object).

    `dimensions` is a pair (2-tuple) of (rows_dimension, columns_dimension) Dimension
    objects. These are always the last two dimensions of `cube` but may and often do
    have transformations applied that are not present on the `cube` dimensions from
    which they derive.

    `slice_idx` is an int offset indicating which portion of `cube` data to use for this
    matrix. There is one slice for each element of the first cube dimension (the "table"
    dimension) when the cube has more than two dimensions.
    """

    def __init__(self, cube, dimensions, slice_idx):
        self._cube = cube
        self._dimensions = dimensions
        self._slice_idx = slice_idx

    @lazyproperty
    def column_aliases(self):
        """1D str ndarray of column alias for each matrix row.

        These are suitable for use as column headings; alias for subtotal columns appear
        in the sequence and alias are ordered to correspond with their respective data
        column.
        """
        return self._dimension_aliases(self._columns_dimension, self._column_order)

    @lazyproperty
    def column_codes(self):
        """1D int ndarray of column codes for each matrix row.

        These are suitable for use as column headings; codes for subtotal columns appear
        in the sequence and codes are ordered to correspond with their respective data
        column.
        """
        return self._dimension_codes(self._columns_dimension, self._column_order)

    @lazyproperty
    def column_comparable_counts(self):
        """2D np.float64 ndarray of counts comparable in the column dimension.

        Comparable in the "column" dimension means that it's not a column subtotal
        difference because these are not considered comparable since their bases
        aren't the same. The values along these subtotal differences are defined as
        np.nan.
        """
        return self._assemble_matrix(self._measures.column_comparable_counts.blocks)

    @lazyproperty
    def column_index(self):
        """2D np.float64 ndarray of column-index "percentage" for each table cell."""
        return self._assemble_matrix(self._measures.column_index.blocks)

    @lazyproperty
    def column_labels(self):
        """1D str ndarray of column name for each matrix column.

        These are suitable for use as column headings; labels for subtotal columns
        appear in the sequence and labels are ordered to correspond with their
        respective data column.
        """
        return self._dimension_labels(self._columns_dimension, self._column_order)

    @lazyproperty
    def column_proportions(self):
        """2D np.float64 ndarray of column-proportion for each matrix cell.

        This is the proportion of the weighted-count for cell to the weighted-N of the
        column the cell appears in (aka. column-margin). Generally a number between 0.0
        and 1.0 inclusive, but subtotal differences can be between -1.0 and 1.0
        inclusive.
        """
        return self._assemble_matrix(self._measures.column_proportions.blocks)

    @lazyproperty
    def column_proportion_variances(self):
        """2D ndarray of np.float64 column-proportion variance for each matrix cell."""
        return self._assemble_matrix(self._measures.column_proportion_variances.blocks)

    @lazyproperty
    def column_share_sum(self):
        """2D optional np.float64 ndarray of column share of sum for each cell.

        Raises `ValueError` if the cube-result does not include a sum cube-measure.
        """
        return self._assemble_matrix(self._measures.column_share_sum.blocks)

    @lazyproperty
    def column_std_err(self):
        """2D np.float64 ndarray of std-error of column-percent for each matrix cell."""
        return self._assemble_matrix(self._measures.column_std_err.blocks)

    @lazyproperty
    def column_unweighted_bases(self):
        """2D np.float64 ndarray of unweighted col-proportions denominator per cell."""
        return self._assemble_matrix(self._measures.column_unweighted_bases.blocks)

    @lazyproperty
    def column_weighted_bases(self):
        """2D np.float64 ndarray of column-proportions denominator for each cell."""
        return self._assemble_matrix(self._measures.column_weighted_bases.blocks)

    @lazyproperty
    def columns_base(self):
        """1D/2D np.float64 ndarray of unweighted-N for each slice column/cell."""
        # --- an MR_X slice produces a 2D columns-base (each cell has its own N) ---
        # --- This is really just another way to call the columns_weighted_bases ---
        # --- TODO: Should column_base only be defined when it's 1D? This would
        # --- require changes to exporter to use the bases to give a
        # --- "column_base_range"
        if not self._measures.columns_unweighted_base.is_defined:
            return self.column_unweighted_bases

        # --- otherwise columns-base is a vector ---
        return self._assemble_marginal(self._measures.columns_unweighted_base)

    @lazyproperty
    def columns_dimension_numeric_values(self):
        """1D optional np.int/float64 ndarray of numeric-value for each column element.

        A value of np.nan appears for a column element without a numeric-value. All
        subtotal columns have a value of np.nan (subtotals have no numeric value).
        """
        elements = self._columns_dimension.valid_elements
        return np.array(
            [
                (elements[idx].numeric_value if idx >= 0 else np.nan)
                for idx in self._column_order
            ]
        )

    @lazyproperty
    def columns_margin(self):
        """1D/2D np.float64 ndarray of weighted-N for each column of this slice."""
        # --- an MR_X slice produces a 2D columns-margin (each cell has its own N) ---
        # --- This is really just another way to call the columns_weighted_bases ---
        # --- TODO: Should column_margin only be defined when it's 1D? This would
        # --- require changes to exporter to use the bases to give a
        # --- "column_margin_range"
        if not self._measures.columns_weighted_base.is_defined:
            return self.column_weighted_bases

        # --- otherwise columns-base is a vector ---
        return self._assemble_marginal(self._measures.columns_weighted_base)

    @lazyproperty
    def columns_margin_proportion(self):
        """1D/2D np.float64 ndarray of weighted-prop for each column of this slice."""
        # --- an MR_X slice produces a 2D columns-margin (each cell has its own N) ---
        # --- TODO: Should colums_margin_proportion only be defined when it's 1D? This
        # --- requires changes to exporter to use the bases to give a
        # --- "columns_margin_range"
        if not self._measures.columns_table_proportion.is_defined:
            return self._assemble_matrix(
                SumSubtotals.blocks(
                    self.columns_margin / self.table_weighted_bases,
                    self._dimensions,
                )
            )

        # --- otherwise columns-margin-proportion is a maginal ---
        return self._assemble_marginal(self._measures.columns_table_proportion)

    @lazyproperty
    def columns_scale_mean(self):
        """Optional 1D column scale mean marginal for this cube-result"""
        return self._assemble_marginal(self._measures.columns_scale_mean)

    @lazyproperty
    def columns_scale_mean_stddev(self):
        """Optional 1D column scale mean std dev marginal for this cube-result"""
        return self._assemble_marginal(self._measures.columns_scale_mean_stddev)

    @lazyproperty
    def columns_scale_mean_stderr(self):
        """Optional 1D column scale mean std err marginal for this cube-result"""
        return self._assemble_marginal(self._measures.columns_scale_mean_stderr)

    @lazyproperty
    def columns_scale_median(self):
        """Optional 1D column scale median marginal for this cube-result"""
        return self._assemble_marginal(self._measures.columns_scale_median)

    @lazyproperty
    def derived_column_idxs(self):
        """tuple(int) of derived column elements' indexes, can be empty.

        An element is derived if it's a subvariable of a multiple response dimension,
        which has been produced by the zz9, and inserted into the response data.

        All other elements, including regular MR and CA subvariables, as well as
        categories of CAT dimensions, are not derived. Subtotals are also not derived
        in this sense, because they're not even part of the data (elements).
        """
        return self._derived_element_idxs(self._columns_dimension, self._column_order)

    @lazyproperty
    def derived_row_idxs(self):
        """tuple(int) of derived row elements' indexes, can be empty.

        An element is derived if it's a subvariable of a multiple response dimension,
        which has been produced by the zz9, and inserted into the response data.

        All other elements, including regular MR and CA subvariables, as well as
        categories of CAT dimensions, are not derived. Subtotals are also not derived
        in this sense, because they're not even part of the data (elements).
        """
        return self._derived_element_idxs(self._rows_dimension, self._row_order)

    @lazyproperty
    def diff_row_idxs(self):
        """tuple(int) of difference row elements' indexes, can be empty."""
        return self._diff_element_idxs(self._rows_dimension, self._row_order)

    @lazyproperty
    def diff_column_idxs(self):
        """tuple(int) of difference column elements' indexes, can be empty."""
        return self._diff_element_idxs(self._columns_dimension, self._column_order)

    @lazyproperty
    def inserted_column_idxs(self):
        """tuple of int index of each subtotal column in slice."""
        # --- insertions have a negative idx in their order sequence ---
        return tuple(i for i, col_idx in enumerate(self._column_order) if col_idx < 0)

    @lazyproperty
    def inserted_row_idxs(self):
        """tuple of int index of each subtotal row in slice."""
        # --- insertions have a negative idx in their order sequence ---
        return tuple(i for i, row_idx in enumerate(self._row_order) if row_idx < 0)

    @lazyproperty
    def means(self):
        """2D optional np.float64 ndarray of mean for each cell.

        Raises `ValueError` if the cube-result does not include a means cube-measure.
        """
        return self._assemble_matrix(self._measures.means.blocks)

    def pairwise_indices(self, alpha, only_larger):
        """2D optional ndarray of tuple of int column-idxs pairwise-t threshold.

        Raises `ValueError if the cube-result does not include `overlaps` cube-measures.
        """
        return np.array(
            [
                self._pairwise_indices(
                    self.pairwise_significance_p_vals(col),
                    self.pairwise_significance_t_stats(col),
                    alpha,
                    only_larger,
                )
                for col in range(len(self._column_order))
            ]
        ).T

    def pairwise_means_indices(self, alpha, only_larger):
        """2D optional ndarray of tuple of int column-idxs means pairwise-t threshold.

        Raises `ValueError if the cube-result does not include `means` cube-measures.
        """
        return np.array(
            [
                self._pairwise_indices(
                    self.pairwise_significance_means_p_vals(col),
                    self.pairwise_significance_means_t_stats(col),
                    alpha,
                    only_larger,
                )
                for col in range(len(self._column_order))
            ]
        ).T

    def pairwise_significance_p_vals(self, column_idx):
        """2D optional np.float64 ndarray of overlaps-p_vals matrices for subvar idx.

        For cubes where the last dimension is categorical, column idxs represent
        specific categories.

        For cubes where the last dimension is a multiple response, each subvariable
        pairwise significance matrix is a 2D ndarray of the p-vals for the selected
        subvariable index (the selected column).

        Raises `ValueError if the cube-result does not include `overlaps`
        and `valid_overlaps` cube-measures.
        """
        if self._cube_has_overlaps:
            # If overlaps are defined, calculate significance based on them
            return self._assemble_matrix(
                self._measures.pairwise_p_vals_for_subvar(column_idx).blocks
            )
        base_column_idx = self._column_order[column_idx]
        return self._assemble_matrix(
            self._measures.pairwise_p_vals(base_column_idx).blocks
        )

    def pairwise_significance_t_stats(self, column_idx):
        """2D optional np.float64 ndarray of overlaps-t_stats matrices for subvar idx.

        For cubes where the last dimension is categorical, column idxs represent
        specific categories.

        For cubes where the last dimension is a multiple response, each subvariable
        pairwise significance matrix is a 2D ndarray of the t-stats for the selected
        subvariable index (the selected column).

        Raises `ValueError if the cube-result does not include `overlaps`
        and `valid_overlaps` cube-measures.
        """
        if self._cube_has_overlaps:
            # If overlaps are defined, calculate significance based on them
            return self._assemble_matrix(
                self._measures.pairwise_t_stats_for_subvar(column_idx).blocks
            )
        base_column_idx = self._column_order[column_idx]
        return self._assemble_matrix(
            self._measures.pairwise_t_stats(base_column_idx).blocks
        )

    def pairwise_significance_means_p_vals(self, column_idx):
        """2D optional np.float64 ndarray of mean difference p_vals for column idx.

        Raises `ValueError if the cube-result does not include `mean` cube-measures.
        """
        base_column_idx = self._column_order[column_idx]
        return self._assemble_matrix(
            self._measures.pairwise_significance_means_p_vals(base_column_idx).blocks
        )

    def pairwise_significance_means_t_stats(self, column_idx):
        """2D optional np.float64 ndarray of mean difference t_stats for column idx.

        Raises `ValueError if the cube-result does not include `mean` cube-measures.
        """
        base_column_idx = self._column_order[column_idx]
        return self._assemble_matrix(
            self._measures.pairwise_significance_means_t_stats(base_column_idx).blocks
        )

    @lazyproperty
    def population_proportions(self):
        """2D np.float64 ndarray of proportions

        The proportion used to calculate proportion counts depends on the dimension
        types.
        """
        return self._assemble_matrix(self._measures.population_proportions.blocks)

    @lazyproperty
    def population_std_err(self):
        """2D np.float64 ndarray of standard errors

        The proportion used to calculate proportion counts depends on the dimension
        types.
        """
        return self._assemble_matrix(self._measures.population_std_err.blocks)

    @lazyproperty
    def pvalues(self):
        """2D np.float64/np.nan ndarray of p-value for each matrix cell."""
        return self._assemble_matrix(self._measures.pvalues.blocks)

    @lazyproperty
    def row_comparable_counts(self):
        """2D np.float64 ndarray of counts comparable in the row dimension.

        Comparable in the "row" dimension means that it's not a row subtotal
        difference because these are not considered comparable since their bases
        aren't the same. The values along these subtotal differences are defined as
        np.nan.
        """
        return self._assemble_matrix(self._measures.row_comparable_counts.blocks)

    @lazyproperty
    def row_aliases(self):
        """1D str ndarray of row alias for each matrix row.

        These are suitable for use as row headings; alias for subtotal rows appear in
        the sequence and alias are ordered to correspond with their respective data
        row.
        """
        return self._dimension_aliases(self._rows_dimension, self._row_order)

    @lazyproperty
    def row_codes(self):
        """1D int ndarray of row codes for each matrix row.

        These are suitable for use as row headings; codes for subtotal rows appear in
        the sequence and codes are ordered to correspond with their respective data
        row.
        """
        return self._dimension_codes(self._rows_dimension, self._row_order)

    @lazyproperty
    def row_labels(self):
        """1D str ndarray of row name for each matrix row.

        These are suitable for use as row headings; labels for subtotal rows appear in
        the sequence and labels are ordered to correspond with their respective data
        row.
        """
        return self._dimension_labels(self._rows_dimension, self._row_order)

    @lazyproperty
    def row_proportions(self):
        """2D np.float64 ndarray of row-proportion for each matrix cell.

        This is the proportion of the weighted-count for cell to the weighted-N of the
        row the cell appears in (aka. row-margin). Generally a number between 0.0 and
        1.0 inclusive, but subtotal differences can be between -1.0 and 1.0 inclusive.
        """
        return self._assemble_matrix(self._measures.row_proportions.blocks)

    @lazyproperty
    def row_proportion_variances(self):
        """2D ndarray of np.float64 row-proportion variance for each matrix cell."""
        return self._assemble_matrix(self._measures.row_proportion_variances.blocks)

    @lazyproperty
    def row_share_sum(self):
        """2D optional np.float64 ndarray of row share of sum for each cell.

        Raises `ValueError` if the cube-result does not include a sum cube-measure.
        """
        return self._assemble_matrix(self._measures.row_share_sum.blocks)

    @lazyproperty
    def row_std_err(self):
        """2D np.float64 ndarray of std-error of row-percent for each matrix cell."""
        return self._assemble_matrix(self._measures.row_std_err.blocks)

    @lazyproperty
    def row_unweighted_bases(self):
        """2D np.float64 ndarray of unweighted row-proportions denominator per cell."""
        return self._assemble_matrix(self._measures.row_unweighted_bases.blocks)

    @lazyproperty
    def row_weighted_bases(self):
        """2D np.float64 ndarray of row-proportions denominator for each cell."""
        return self._assemble_matrix(self._measures.row_weighted_bases.blocks)

    @lazyproperty
    def rows_base(self):
        """1D/2D np.float64 ndarray of unweighted-N for each slice row/cell."""
        # --- an X_ARRAY slice produces a 2D row-base (each cell has its own N) ---
        # --- TODO: Should rows_base only be defined when it's 1D? This would
        # --- require changes to exporter to use the bases to give a "rows_base_range"
        if not self._measures.rows_unweighted_base.is_defined:
            return self.row_unweighted_bases

        # --- otherwise rows-base is a vector ---
        return self._assemble_marginal(self._measures.rows_unweighted_base)

    @lazyproperty
    def rows_dimension_fills(self):
        """tuple of RGB str like "#def032" fill color for each row in slice."""
        elements = self._rows_dimension.valid_elements
        subtotals = self._rows_dimension.subtotals
        return tuple(
            # ---Subtotals have negative sequential indexes (-1, -2, ..., -m)---
            # ---To index them properly, we need to convert those indexes to---
            # ---zero based positive indexes (0, 1, ... m - 1) i.e. -idx - 1---
            (elements[idx].fill if idx >= 0 else subtotals[idx + len(subtotals)].fill)
            for idx in self._row_order
        )

    @lazyproperty
    def rows_dimension_numeric_values(self):
        """1D optional np.int/float64 ndarray of numeric-value for each row element.

        A value of np.nan appears for a row element without a numeric-value. All
        subtotal rows have a value of np.nan (subtotals have no numeric value).
        """
        elements = self._rows_dimension.valid_elements
        return np.array(
            [
                (elements[idx].numeric_value if idx >= 0 else np.nan)
                for idx in self._row_order
            ]
        )

    @lazyproperty
    def rows_margin(self):
        """1D/2D np.float64 ndarray of weighted-N for each column of this slice.

        This value is a 2D array if one or both of the slice dimensions are MR. This is
        because each cell has a distinct margin in such a slice.
        """
        # --- an X_MR slice produces a 2D rows-margin (each cell has its own N) ---
        # --- This is really just another way to call the row_weighted_bases ---
        # --- TODO: Should rows_margin only be defined when it's 1D? This would
        # --- require changes to exporter to use the bases to give a "rows_margin_range"
        if not self._measures.rows_weighted_base.is_defined:
            return self.row_weighted_bases

        # --- otherwise rows-margin is a vector ---
        return self._assemble_marginal(self._measures.rows_weighted_base)

    @lazyproperty
    def rows_margin_proportion(self):
        """1D/2D np.float64 ndarray of weighted-proportion for each slice row/cell."""
        # --- an X_MR slice produces a 2D rows-margin (each cell has its own N) ---
        # --- TODO: Should rows_margin_proportion only be defined when it's 1D? This would
        # --- require changes to exporter to use the bases to give a "rows_margin_range"
        if not self._measures.rows_table_proportion.is_defined:
            return self._assemble_matrix(
                SumSubtotals.blocks(
                    self.rows_margin / self.table_weighted_bases,
                    self._dimensions,
                )
            )

        # --- otherwise rows-margin is a vector ---
        return self._assemble_marginal(self._measures.rows_table_proportion)

    @lazyproperty
    def rows_scale_mean(self):
        """Optional 1D Row scale median marginal for this cube-result"""
        return self._assemble_marginal(self._measures.rows_scale_mean)

    @lazyproperty
    def rows_scale_mean_stddev(self):
        """Optional 1D row scale mean std dev marginal for this cube-result"""
        return self._assemble_marginal(self._measures.rows_scale_mean_stddev)

    @lazyproperty
    def rows_scale_mean_stderr(self):
        """Optional 1D row scale mean std dev marginal for this cube-result"""
        return self._assemble_marginal(self._measures.rows_scale_mean_stderr)

    @lazyproperty
    def rows_scale_median(self):
        """Optional 1D Row scale median marginal for this cube-result"""
        return self._assemble_marginal(self._measures.rows_scale_median)

    @lazyproperty
    def smoothed_column_index(self):
        """2D np.float64 ndarray of smoothed column-index for each table cell."""
        return self._assemble_matrix(self._measures.smoothed_column_index.blocks)

    @lazyproperty
    def smoothed_column_proportions(self):
        """2D np.float64 ndarray of smoothed column-proportion for each matrix cell.

        This is the proportion of the weighted-count for cell to the weighted-N of the
        column the cell appears in (aka. column-margin). Generally a number between 0.0
        and 1.0 inclusive, but subtotal differences can be between -1.0 and 1.0
        inclusive.
        """
        return self._assemble_matrix(self._measures.smoothed_column_proportions.blocks)

    @lazyproperty
    def smoothed_columns_scale_mean(self):
        """Optional 1D smoothed column scale mean marginal for this cube-result"""
        return self._assemble_marginal(self._measures.smoothed_columns_scale_mean)

    @lazyproperty
    def smoothed_means(self):
        """2D optional np.float64 ndarray of smoothed mean for each cell.

        Raises `ValueError` if the cube-result does not include a means cube-measure.
        """
        return self._assemble_matrix(self._measures.smoothed_means.blocks)

    @lazyproperty
    def sums(self):
        """2D optional np.float64 ndarray of sum for each cell.

        Raises `ValueError` if the cube-result does not include a sum cube-measure.
        """
        return self._assemble_matrix(self._measures.sums.blocks)

    @lazyproperty
    def stddev(self):
        """2D optional np.float64 ndarray of stddev for each cell.

        Raises `ValueError` if the cube-result does not include a stddev cube-measure.
        """
        return self._assemble_matrix(self._measures.stddev.blocks)

    @lazyproperty
    def table_base(self):
        """Scalar, 1D, or 2D ndarray of np.float64 unweighted-N for this slice.

        This value has four distinct forms, depending on the slice dimensions:

            * ARR_X_ARR - 2D ndarray with a distinct table-base value per cell.
            * ARR_X - 1D ndarray of value per *row* when only rows dimension is ARR.
            * X_ARR - 1D ndarray of value per *column* when only columns dimension is ARR
            * CAT_X_CAT - scalar float value when slice has no MR dimension.

        """
        if self._measures.table_unweighted_base.is_defined:
            return self._measures.table_unweighted_base.value
        if self._measures.columns_table_unweighted_base.is_defined:
            return self._assemble_marginal(self._measures.columns_table_unweighted_base)
        if self._measures.rows_table_unweighted_base.is_defined:
            return self._assemble_marginal(self._measures.rows_table_unweighted_base)
        return self.table_unweighted_bases

    @lazyproperty
    def table_base_range(self):
        """[min, max] np.float64 ndarray range of the table_margin (table-weighted-base)

        A CAT_X_CAT has a scalar for all table-weighted-bases, but arrays have more than
        one table-weighted-base. This collapses all the values them to the range, and
        it is "unpruned", meaning that it is calculated before any hiding or removing
        of empty rows/columns.
        """
        return self._measures.table_unweighted_bases_range.value

    @lazyproperty
    def table_margin(self):
        """Scalar, 1D, or 2D ndarray of np.float64 weighted-N for this slice.

        This value has four distinct forms, depending on the slice dimensions:

            * CAT_X_CAT - scalar float value when slice has no ARRAY dimension.
            * ARRAY_X - 1D ndarray of value per *row* when only rows dimension is ARRAY.
            * X_ARRAY - 1D ndarray of value per *column* when only column is ARRAY.
            * ARRAY_X_ARRAY - 2D ndarray with a distinct table-margin value per cell.
        """
        if self._measures.table_weighted_base.is_defined:
            return self._measures.table_weighted_base.value
        if self._measures.columns_table_weighted_base.is_defined:
            return self._assemble_marginal(self._measures.columns_table_weighted_base)
        if self._measures.rows_table_weighted_base.is_defined:
            return self._assemble_marginal(self._measures.rows_table_weighted_base)
        return self.table_weighted_bases

    @lazyproperty
    def table_proportion_variances(self):
        """2D ndarray of np.float64 table-proportion variance for each matrix cell."""
        return self._assemble_matrix(self._measures.table_proportion_variances.blocks)

    @lazyproperty
    def table_proportions(self):
        """2D np.float64 ndarray of table-proportion for each matrix cell.

        This is the proportion of the weighted-count for cell to the weighted-N of the
        row the cell appears in (aka. table-margin). Generally a number between 0.0 and
        1.0 inclusive, but subtotal differences can be between -1.0 and 1.0 inclusive.
        """
        return self._assemble_matrix(self._measures.table_proportions.blocks)

    @lazyproperty
    def table_std_err(self):
        """2D np.float64 ndarray of std-error of table-percent for each matrix cell."""
        return self._assemble_matrix(self._measures.table_std_err.blocks)

    @lazyproperty
    def table_unweighted_bases(self):
        """2D np.float64 ndarray of unweighted table-proportion denominator per cell."""
        return self._assemble_matrix(self._measures.table_unweighted_bases.blocks)

    @lazyproperty
    def table_weighted_bases(self):
        """2D np.float64 ndarray of weighted table-proportion denominator per cell."""
        return self._assemble_matrix(self._measures.table_weighted_bases.blocks)

    @lazyproperty
    def total_share_sum(self):
        """2D optional np.float64 ndarray of total share of sum for each cell.

        Raises `ValueError` if the cube-result does not include a sum
        cube-measure.
        """
        return self._assemble_matrix(self._measures.total_share_sum.blocks)

    @lazyproperty
    def table_margin_range(self):
        """[min, max] np.float64 ndarray range of the table_margin (table-weighted-base)

        A CAT_X_CAT has a scalar for all table-weighted-bases, but arrays have more than
        one table-weighted-base. This collapses all of the values to a range, and
        it is "unpruned", meaning that it is calculated before any hiding or removing
        of empty rows/columns.
        """
        return self._measures.table_weighted_bases_range.value

    @lazyproperty
    def unweighted_counts(self):
        """2D np.float64 ndarray of unweighted-count for each cell."""
        return self._assemble_matrix(self._measures.unweighted_counts.blocks)

    @lazyproperty
    def weighted_counts(self):
        """2D np.float64 ndarray of weighted-count for each cell."""
        return self._assemble_matrix(self._measures.weighted_counts.blocks)

    @lazyproperty
    def zscores(self):
        """2D np.float64 ndarray of std-res value for each cell of matrix.

        A z-score is also known as a *standard score* and is the number of standard
        deviations above (positive) or below (negative) the population mean a cell's
        value is.
        """
        return self._assemble_matrix(self._measures.zscores.blocks)

    def _assemble_marginal(self, marginal):
        """Optional 1D ndarray created from a marginal.

        The assembled marginal is the shape of either a row or column (determined by
        `marginal.orientation`), and with the ordering that's applied to those
        dimensions.

        It is None when the marginal is not defined (`marginal._is_defined`).
        """
        if not marginal.is_defined:
            return None

        order = (
            self._row_order if marginal.orientation == MO.ROWS else self._column_order
        )

        return np.hstack(marginal.blocks)[order]

    def _assemble_matrix(self, blocks):
        """Return 2D ndarray matrix assembled from `blocks`.

        The assembled matrix includes inserted vectors (rows and columns), has hidden
        vectors removed, and is ordered by whatever sort method is applied in the
        dimension transforms.
        """
        # --- These are assembled into a single 2D array, and then rearranged based on
        # --- row and column orders. All insertion, ordering, and hiding transforms are
        # --- reflected in the row and column orders. They each include (negative)
        # --- insertion idxs, hidden and pruned vector indices have been removed, and
        # --- the ordering method has been applied to determine the sequence each idx
        # --- appears in. This directly produces a final array that is exactly the
        # --- desired output.
        return np.block(blocks)[np.ix_(self._row_order, self._column_order)]

    def _assemble_vector(self, base_vector, subtotals, order, diffs_nan=False):
        """Return 1D ndarray of `base_vector` with inserted `subtotals`, in `order`.

        Each subtotal value is the result of applying np.sum to the addends and
        subtrahends extracted from `base_vector` according to the `addend_idxs`
        and `subtrahend_idxs` property of each subtotal in `subtotals`. The returned
        array is arranged by `order`, including possibly removing hidden or pruned
        values.
        """
        # TODO: This works for "sum" and "diff" subtotals, because either we set to
        # nan or add & subtract, but a fuller solution will probably get the subtotal
        # values from a _BaseSubtotals subclass.
        vector_subtotals = np.array(
            [
                np.nan
                if diffs_nan and len(subtotal.subtrahend_idxs) > 0
                else np.sum(base_vector[subtotal.addend_idxs])
                - np.sum(base_vector[subtotal.subtrahend_idxs])
                for subtotal in subtotals
            ]
        )
        return np.hstack([base_vector, vector_subtotals])[order]

    @lazyproperty
    def _column_order(self):
        """1D np.int64 ndarray of signed int idx for each assembled column.

        Negative values represent inserted subtotal-column locations.
        """
        return _BaseOrderHelper.column_display_order(self._dimensions, self._measures)

    @lazyproperty
    def _column_subtotals(self):
        """Sequence of _Subtotal object for each inserted column."""
        return self._columns_dimension.subtotals

    @lazyproperty
    def _columns_dimension(self):
        """The `Dimension` object representing column elements in this matrix."""
        return self._dimensions[1]

    @lazyproperty
    def _cube_has_overlaps(self):
        """True if overlaps are defined and the last dimension is MR, False otherwise"""
        return (
            self._dimensions[-1].dimension_type == DT.MR
            and self._cube.overlaps is not None
            and self._cube.valid_overlaps is not None
        )

    def _derived_element_idxs(self, dimension, order):
        """Return tuple(int) of derived elements' indices for a dimension.

        Subtotals cannot be derived elements. Only some elements (subvariables) can.
        """
        n_subtotals = len(dimension.valid_elements)
        derivs = [e.derived for e in dimension.valid_elements] + [False] * n_subtotals
        return tuple(np.where(np.array(derivs)[order])[0])

    def _diff_element_idxs(self, dimension, order):
        """Return tuple(int) of difference elements' indices for a dimension.

        Valid elements cannot be differences. Only some subtotals can.
        """
        n_valids = len(dimension.valid_elements)
        diffs = [False] * n_valids + [e.is_difference for e in dimension.subtotals]
        return tuple(np.where(np.array(diffs)[order])[0])

    def _dimension_aliases(self, dimension, order):
        """1D str ndarray of alias for each vector of `dimension`.

        These are suitable for use as row/column headings; alias for subtotals appear
        in the sequence specified by `order`, such that aliases are ordered to
        correspond with their respective data vector.
        """
        return np.array(dimension.element_aliases + dimension.subtotal_aliases)[order]

    def _dimension_codes(self, dimension, order):
        """1D str ndarray of codes for each vector of `dimension`.

        These are suitable for use as row/column headings; codes for subtotals appear
        in the sequence specified by `order`, such that codes are ordered to
        correspond with their respective data vector.
        """
        return np.array(dimension.element_ids + dimension.insertion_ids)[order]

    def _dimension_labels(self, dimension, order):
        """1D str ndarray of name for each vector of `dimension`.

        These are suitable for use as row/column headings; labels for subtotals appear
        in the sequence specified by `order`, such that labels are ordered to correspond
        with their respective data vector.
        """
        return np.array(dimension.element_labels + dimension.subtotal_labels)[order]

    @lazyproperty
    def _measures(self):
        """SecondOrderMeasures collection object for this cube-result."""
        return SecondOrderMeasures(self._cube, self._dimensions, self._slice_idx)

    @staticmethod
    def _pairwise_indices(p_vals, t_stats, alpha, only_larger):
        """1D ndarray containing tuples of int pairwise indices of each column."""
        significance = p_vals < alpha
        if only_larger:
            significance = np.logical_and(t_stats < 0, significance)
        col_significance = np.empty((len(significance),), dtype=object)
        col_significance[:] = [tuple(np.where(sig_row)[0]) for sig_row in significance]
        return col_significance

    @lazyproperty
    def _row_order(self):
        """1D np.int64 ndarray of signed int idx for each assembled row.

        Negative values represent inserted subtotal-row locations.
        """
        return _BaseOrderHelper.row_display_order(self._dimensions, self._measures)

    @lazyproperty
    def _row_subtotals(self):
        """Sequence of _Subtotal object for each inserted row."""
        return self._rows_dimension.subtotals

    @lazyproperty
    def _rows_dimension(self):
        """The `Dimension` object representing row elements in this matrix."""
        return self._dimensions[0]


class _BaseOrderHelper:
    """Base class for ordering helpers."""

    def __init__(self, dimensions, second_order_measures):
        self._dimensions = dimensions
        self._second_order_measures = second_order_measures

    @classmethod
    def column_display_order(cls, dimensions, second_order_measures):
        """1D np.int64 ndarray of signed int idx for each column of measure matrix.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        columns-dimension.
        """
        # --- This is essentially a factory method. There is no sort-columns-by-value
        # --- yet, and both explicit and payload ordering are handled by
        # --- _ColumnOrderHelper, so there's not much to this yet, just keeping
        # --- form consistent with `.row_display_order()` and we'll elaborate this when
        # --- we add sort-by-value to columns.
        return _ColumnOrderHelper(dimensions, second_order_measures)._display_order

    @classmethod
    def row_display_order(cls, dimensions, second_order_measures):
        """1D np.int64 ndarray of signed int idx for each row of measure matrix.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        collation_method = dimensions[0].order_spec.collation_method
        dim_type = dimensions[1].dimension_type
        HelperCls = (
            _SortRowsByBaseColumnHelper
            if collation_method == CM.OPPOSING_ELEMENT
            else _SortRowsByDerivedColumnHelper
            if collation_method == CM.OPPOSING_INSERTION and dim_type in DT.ARRAY_TYPES
            else _SortRowsByInsertedColumnHelper
            if collation_method == CM.OPPOSING_INSERTION
            else _SortRowsByLabelHelper
            if collation_method == CM.LABEL
            else _SortRowsByMarginalHelper
            if collation_method == CM.MARGINAL
            else _RowOrderHelper
        )

        return HelperCls(dimensions, second_order_measures)._display_order

    @lazyproperty
    def _display_order(self):
        """1D np.int64 ndarray of signed int idx for each vector of dimension.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        # --- Returning as np.array suits its intended purpose, which is to participate
        # --- in an np._ix() call. It works fine as a sequence too for any alternate
        # --- use. Specifying int type prevents failure when there are zero elements.
        if self._prune_subtotals:
            return np.array([idx for idx in self._order if idx >= 0], dtype=int)
        return np.array(self._order, dtype=int)

    @lazyproperty
    def _columns_dimension(self):
        """The `Dimension` object representing column elements in the matrix."""
        return self._dimensions[1]

    @lazyproperty
    def _empty_column_idxs(self):
        """tuple of int index for each column with (unweighted) N = 0.

        These columns are subject to pruning, depending on a user setting in the
        dimension.
        """
        # --- subset because numpy.where returns tuple to allow for multiple axes
        return tuple(np.where(self._second_order_measures.columns_pruning_mask)[0])

    @lazyproperty
    def _empty_row_idxs(self):
        """tuple of int index for each row with N = 0.

        These rows are subject to pruning, depending on a user setting in the dimension.
        """
        # --- subset because numpy.where returns tuple to allow for multiple axes
        return tuple(np.where(self._second_order_measures.rows_pruning_mask)[0])

    @lazyproperty
    def _measure(self):
        """Second-order measure object providing values for sort.

        Returns None when the measure in the `order_spec` is not enable for sorting.

        This property is not used by some subclasses.
        """
        # --- Sometimes we sort by a measure that is a monotonic transformation of
        # --- another measure that is already calculated by the assembler. Because
        # --- monotonic transformations preserve order, the sorting is correct.
        propname_by_measure = {
            M.COLUMN_BASE_UNWEIGHTED: "column_unweighted_bases",
            M.COLUMN_BASE_WEIGHTED: "column_weighted_bases",
            M.COLUMN_INDEX: "column_index",
            M.COLUMN_PERCENT: "column_proportions",
            M.COLUMN_PERCENT_MOE: "column_std_err",  # monotonic transform
            M.COLUMN_SHARE_SUM: "column_share_sum",
            M.COLUMN_STDDEV: "column_proportion_variances",  # monotonic transform
            M.COLUMN_STDERR: "column_std_err",
            M.MEAN: "means",
            M.POPULATION: "population_proportions",  # montonic transform
            M.POPULATION_MOE: "population_std_err",  # monotonic transform
            M.PVALUES: "pvalues",
            M.ROW_BASE_UNWEIGHTED: "row_unweighted_bases",
            M.ROW_BASE_WEIGHTED: "row_weighted_bases",
            M.ROW_PERCENT: "row_proportions",
            M.ROW_PERCENT_MOE: "row_std_err",  # monotonic transform
            M.ROW_SHARE_SUM: "row_share_sum",
            M.ROW_STDDEV: "row_proportion_variances",  # montonic transform
            M.ROW_STDERR: "row_std_err",
            M.STDDEV: "stddev",
            M.SUM: "sums",
            M.TABLE_PERCENT: "table_proportions",
            M.TABLE_PERCENT: "table_proportions",
            M.TABLE_PERCENT_MOE: "table_std_err",  # monotonic transform
            M.TABLE_STDDEV: "table_proportion_variances",  # monotonic transform
            M.TABLE_STDERR: "table_std_err",
            M.TABLE_BASE_UNWEIGHTED: "table_unweighted_bases",
            M.TABLE_BASE_WEIGHTED: "table_weighted_bases",
            M.TOTAL_SHARE_SUM: "total_share_sum",
            M.UNWEIGHTED_COUNT: "unweighted_counts",
            M.UNWEIGHTED_VALID_COUNT: "unweighted_counts",
            M.WEIGHTED_COUNT: "weighted_counts",
            M.WEIGHTED_VALID_COUNT: "weighted_counts",
            M.Z_SCORE: "zscores",
        }
        measure = self._order_spec.measure
        measure_propname = propname_by_measure.get(measure)

        if measure_propname is None:
            raise NotImplementedError(
                f"sort-by-value for measure '{measure}' is not yet supported"
            )

        return getattr(self._second_order_measures, measure_propname)

    @lazyproperty
    def _order(self):
        """tuple of signed int idx for each sorted vector of measure matrix.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        raise NotImplementedError(  # pragma: no cover
            f"{type(self).__name__} must implement `._order`"
        )

    @lazyproperty
    def _order_spec(self):
        """_OrderSpec object for dimension being sorted.

        Provides access to ordering details like measure and sort-direction.
        """
        raise NotImplementedError(  # pragma: no cover
            f"{type(self).__name__} must implement `._order_spec`"
        )

    @lazyproperty
    def _prune_subtotals(self):
        """True if subtotal vectors need to be pruned, False otherwise."""
        raise NotImplementedError(  # pragma: no cover
            f"{type(self).__name__} must implement `._prune_subtotals`"
        )

    @lazyproperty
    def _rows_dimension(self):
        """The `Dimension` object representing row elements in the matrix."""
        return self._dimensions[0]


class _ColumnOrderHelper(_BaseOrderHelper):
    """Encapsulates the complexity of the various kinds of column ordering."""

    @lazyproperty
    def _order(self):
        """tuple of signed int idx for each column of measure matrix.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        CollatorCls = (
            ExplicitOrderCollator
            if self._order_spec.collation_method == CM.EXPLICIT_ORDER
            else PayloadOrderCollator
        )
        return CollatorCls.display_order(
            self._columns_dimension, self._empty_column_idxs
        )

    @lazyproperty
    def _order_spec(self):
        """_OrderSpec object for the columns dimension.

        Provides access to ordering details like measure and sort-direction.
        """
        return self._columns_dimension.order_spec

    @lazyproperty
    def _prune_subtotals(self):
        """True if subtotal columns need to be pruned, False otherwise.

        Subtotal-columns need to be pruned when all base-rows are pruned. Subtotal
        columns are only subject to pruning when row-pruning is specified in the
        request.
        """
        return (
            len(self._empty_row_idxs) == len(self._rows_dimension.element_ids)
            if self._rows_dimension.prune
            else False
        )


class _RowOrderHelper(_BaseOrderHelper):
    """Encapsulates the complexity of the various kinds of row ordering."""

    @lazyproperty
    def _order(self):
        """tuple of signed int idx for each row of measure matrix.

        Negative values represent inserted-vector locations. Returned sequence reflects
        insertion, hiding, pruning, and ordering transforms specified in the
        rows-dimension.
        """
        CollatorCls = (
            ExplicitOrderCollator
            if self._order_spec.collation_method == CM.EXPLICIT_ORDER
            else PayloadOrderCollator
        )
        return CollatorCls.display_order(self._rows_dimension, self._empty_row_idxs)

    @lazyproperty
    def _order_spec(self):
        """_OrderSpec object for this row dimension.

        Provides access to ordering details like measure and sort-direction.
        """
        return self._rows_dimension.order_spec

    @lazyproperty
    def _prune_subtotals(self):
        """True if subtotal rows need to be pruned, False otherwise.

        Subtotal-rows need to be pruned when all base-columns are pruned. Subtotal rows
        only subject to pruning when column-pruning is specified in the request.
        """
        return (
            len(self._empty_column_idxs) == len(self._columns_dimension.element_ids)
            if self._columns_dimension.prune
            else False
        )


class _BaseSortRowsByValueHelper(_RowOrderHelper):
    """A base class that orders elements by a set of values.

    This class is intended only to serve as a base for the other sort-by-value classes
    which must all provide their own implentations of `_element_values` and
    `_subtotal_values`.

    If `_order` encouters a ValueError, it falls back to the payload order.
    """

    @lazyproperty
    def _element_values(self):
        """Sequence of body values that form the basis for sort order.

        Must be implemented by child classes.
        """
        raise NotImplementedError(  # pragma: no cover
            f"{type(self).__name__} must implement `._element_values`"
        )

    @lazyproperty
    def _order(self):
        """tuple of int element-idx specifying ordering of dimension elements."""
        try:
            return SortByValueCollator.display_order(
                self._rows_dimension,
                self._element_values,
                self._subtotal_values,
                self._empty_row_idxs,
            )
        except ValueError:
            return PayloadOrderCollator.display_order(
                self._rows_dimension, self._empty_row_idxs
            )

    @lazyproperty
    def _subtotal_values(self):
        """Sequence of subtotal values that form the basis for sort order.

        Must be implemented by child classes.
        """
        raise NotImplementedError(  # pragma: no cover
            f"{type(self).__name__} must implement `._subtotal_values`"
        )


class _SortRowsByBaseColumnHelper(_BaseSortRowsByValueHelper):
    """Orders elements by the values of an opposing base (not a subtotal) vector.

    This would be like "order rows in descending order by value of 'Strongly Agree'
    column. An opposing-element ordering is only available on a matrix, because only
    a matrix dimension has an opposing dimension.
    """

    @lazyproperty
    def _column_idx(self):
        """int index of column whose values the sort is based on."""
        column_element_ids = self._columns_dimension.element_ids
        sort_column_id = self._order_spec.element_id
        # --- Need to translate the element id to the shimmed element id
        sort_column_id = self._columns_dimension.translate_element_id(sort_column_id)
        return column_element_ids.index(sort_column_id)

    @lazyproperty
    def _element_values(self):
        """Sequence of body values that form the basis for sort order.

        There is one value per row and values appear in payload (dimension) element
        order. These are only the "base" values and do not include insertions.
        """
        measure_base_values = self._measure.blocks[0][0]
        return measure_base_values[:, self._column_idx]

    @lazyproperty
    def _subtotal_values(self):
        """Sequence of row-subtotal values that contribute to the sort basis.

        There is one value per row subtotal and values appear in payload (dimension)
        insertion order.
        """
        measure_subtotal_rows = self._measure.blocks[1][0]
        return measure_subtotal_rows[:, self._column_idx]


class _SortRowsByDerivedColumnHelper(_SortRowsByBaseColumnHelper):
    """Orders elements by the values of an opposing "derived" element vector.

    "Derived" columns are insertions that are calculated by zz9 and put into the
    cube result before this cr.cube library gets them. These insertions are always
    on a subvariable dimension, an example would be to sort "'Coke' or 'Diet Coke'"
    MR array column. An opposing-element ordering is only available on a matrix,
    because only a matrix dimension has an opposing dimension.

    For the most part, we treat these elements as regular "base" columns, in the cr.cube
    library, but because the user-facing language treats these as subtotals, we allow
    specifying sort-by-value as if these were true subtotals.
    """

    @lazyproperty
    def _column_idx(self):
        """int index of column whose values the sort is based on."""
        column_element_ids = self._columns_dimension.element_ids
        sort_column_id = self._order_spec.insertion_id
        # --- Need to translate the element id to the shimmed element id
        sort_column_id = self._columns_dimension.translate_element_id(sort_column_id)
        return column_element_ids.index(sort_column_id)


class _SortRowsByInsertedColumnHelper(_BaseSortRowsByValueHelper):
    """Orders rows by the values in an inserted column.

    This would be like "order rows in descending order by value of 'Top 3' subtotal
    column. An opposing-insertion ordering is only available on a matrix because only
    a matrix dimension has an opposing dimension.
    """

    @lazyproperty
    def _element_values(self):
        """Sequence of insertion "body" values that form the basis for sort order.

        There is one value per base row and values appear in payload order. These are
        only the "base" insertion values and do not include intersections.
        """
        insertion_base_values = self._measure.blocks[0][1]
        return insertion_base_values[:, self._insertion_idx]

    @lazyproperty
    def _insertion_idx(self):
        """int index of insertion whose values the sort is based on."""
        return self._columns_dimension.insertion_ids.index(
            self._order_spec.insertion_id
        )

    @lazyproperty
    def _subtotal_values(self):
        """Sequence of row-subtotal intersection values that contribute to sort basis.

        There is one value per row subtotal and values appear in payload (dimension)
        insertion order.
        """
        intersections = self._measure.blocks[1][1]
        return intersections[:, self._insertion_idx]


class _SortRowsByLabelHelper(_BaseSortRowsByValueHelper):
    """Orders elements by the values of their labels (from the dimension)."""

    @lazyproperty
    def _element_values(self):
        """Sequence of body labels that form the basis for sort order.

        There is one value per row and values appear in payload (dimension) element
        order. These are only the "base" values and do not include insertions.
        """
        return np.array(self._dimensions[0].element_labels)

    @lazyproperty
    def _subtotal_values(self):
        """Sequence of row-subtotal labels that contribute to the sort basis.

        There is one value per row subtotal and labels appear in payload (dimension)
        insertion order.
        """
        return np.array(self._dimensions[0].subtotal_labels)


class _SortRowsByMarginalHelper(_BaseSortRowsByValueHelper):
    """Orders elements by the values of an opposing marginal vector.

    This would be like "order rows in descending order by value of rows scale mean.
    column. An opposing-marginal ordering is only available on a matrix, because only
    a matrix dimension has an opposing dimension.
    """

    @lazyproperty
    def _element_values(self):
        """Sequence of body values that form the basis for sort order.

        There is one value per row and values appear in payload (dimension) element
        order. These are only the "base" values and do not include insertions.
        """
        return self._marginal.blocks[0]

    @lazyproperty
    def _marginal(self):
        """Marginal object providing values for sort."""
        marginal = self._order_spec.marginal
        marginal_propname = {
            MARGINAL.BASE: "rows_unweighted_base",
            MARGINAL.MARGIN: "rows_weighted_base",
            MARGINAL.MARGIN_PROPORTION: "rows_table_proportion",
            MARGINAL.SCALE_MEAN: "rows_scale_mean",
            MARGINAL.SCALE_MEAN_STDDEV: "rows_scale_mean_stddev",
            MARGINAL.SCALE_MEAN_STDERR: "rows_scale_mean_stderr",
            MARGINAL.SCALE_MEDIAN: "rows_scale_median",
        }.get(marginal)

        if marginal_propname is None:
            raise NotImplementedError(
                f"sort-by-value for marginal '{marginal}' is not yet supported"
            )

        return getattr(self._second_order_measures, marginal_propname)

    @lazyproperty
    def _subtotal_values(self):
        """Sequence of row-subtotal values that contribute to the sort basis.

        There is one value per row subtotal and values appear in payload (dimension)
        insertion order.
        """
        return self._marginal.blocks[1]



# encoding: utf-8

"""Unit test suite for `cr.cube.stripe.insertion` module."""

import numpy as np
import pytest

from cr.cube.dimension import Dimension, _Subtotal
from cr.cube.stripe.insertion import (
    _BaseSubtotals,
    NanSubtotals,
    NegativeTermSubtotals,
    PositiveTermSubtotals,
    SumSubtotals,
)

from ...unitutil import ANY, initializer_mock, instance_mock, property_mock


class Describe_BaseSubtotals:
    """Unit test suite for `cr.cube.stripe._BaseSubtotals` object."""

    def it_provides_a_subtotal_values_interface_method(self, request, rows_dimension_):
        base_values = [1, 2, 3]
        _init_ = initializer_mock(request, _BaseSubtotals)
        property_mock(
            request, _BaseSubtotals, "_subtotal_values", return_value=np.array([3, 5])
        )

        subtotal_values = _BaseSubtotals.subtotal_values(base_values, rows_dimension_)

        _init_.assert_called_once_with(ANY, base_values, rows_dimension_)
        assert subtotal_values.tolist() == [3, 5]

    def it_provides_access_to_the_row_subtotal_objects(self, request, rows_dimension_):
        row_subtotals_ = tuple(instance_mock(request, _Subtotal) for _ in range(3))
        rows_dimension_.subtotals = row_subtotals_
        subtotals = _BaseSubtotals(None, rows_dimension_)

        assert subtotals._row_subtotals is row_subtotals_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def rows_dimension_(self, request):
        return instance_mock(request, Dimension)


class DescribeNanSubtotals:
    """Unit test suite for `cr.cube.stripe.NanSubtotals` object."""

    def it_computes_the_subtotal_values(self, request):
        property_mock(
            request,
            NanSubtotals,
            "_row_subtotals",
            return_value=tuple(instance_mock(request, _Subtotal) for _ in range(3)),
        )
        nan_subtotals = NanSubtotals(None, None)

        assert nan_subtotals._subtotal_values == pytest.approx(
            [np.nan, np.nan, np.nan], nan_ok=True
        )


class DescribeNegativeTermSubtotals:
    """Unit test suite for `cr.cube.matrix.NegativeTermSubtotals` object."""

    @pytest.mark.parametrize(
        ("addend_idxs", "subtrahend_idxs", "expected"),
        (
            ([0, 1], [2, 3], 12),
            ([1], [], 0),
            ([], [1], 2),
        ),
    )
    def it_computes_the_subtotal_values(
        self, request, addend_idxs, subtrahend_idxs, expected
    ):
        property_mock(
            request,
            _Subtotal,
            "addend_idxs",
            return_value=np.array(addend_idxs, dtype=int),
        )
        property_mock(
            request,
            _Subtotal,
            "subtrahend_idxs",
            return_value=np.array(subtrahend_idxs, dtype=int),
        )

        subtotal_value = NegativeTermSubtotals(
            np.array([1, 2, 4, 8]), None
        )._subtotal_value(_Subtotal(None, None, None))

        np.testing.assert_equal(subtotal_value, expected)


class DescribePositiveTermSubtotals:
    """Unit test suite for `cr.cube.matrix.PositiveTermSubtotals` object."""

    @pytest.mark.parametrize(
        ("addend_idxs", "subtrahend_idxs", "expected"),
        (
            ([0, 1], [2, 3], 3),
            ([1], [], 2),
            ([], [1], 0),
        ),
    )
    def it_computes_the_subtotal_values(
        self, request, addend_idxs, subtrahend_idxs, expected
    ):
        property_mock(
            request,
            _Subtotal,
            "addend_idxs",
            return_value=np.array(addend_idxs, dtype=int),
        )
        property_mock(
            request,
            _Subtotal,
            "subtrahend_idxs",
            return_value=np.array(subtrahend_idxs, dtype=int),
        )

        subtotal_value = PositiveTermSubtotals(
            np.array([1, 2, 4, 8]), None
        )._subtotal_value(_Subtotal(None, None, None))

        np.testing.assert_equal(subtotal_value, expected)


class DescribeSumSubtotals:
    """Unit test suite for `cr.cube.matrix.SumSubtotals` object."""

    @pytest.mark.parametrize(
        ("addend_idxs", "subtrahend_idxs", "expected"),
        (
            ([0, 1], [2, 3], -9),
            ([1], [], 2),
            ([], [1], -2),
        ),
    )
    def it_calculates_subtotal_value_correctly(
        self, request, addend_idxs, subtrahend_idxs, expected
    ):
        property_mock(
            request,
            _Subtotal,
            "addend_idxs",
            return_value=np.array(addend_idxs, dtype=int),
        )
        property_mock(
            request,
            _Subtotal,
            "subtrahend_idxs",
            return_value=np.array(subtrahend_idxs, dtype=int),
        )

        subtotal_value = SumSubtotals(np.array([1, 2, 4, 8]), None)._subtotal_value(
            _Subtotal(None, None, None)
        )

        np.testing.assert_equal(subtotal_value, expected)
